# Copyright (C) 2023 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
import os
import ssl
import sys
from io import BytesIO
from pathlib import Path
from typing import Any, BinaryIO, Dict, List, Optional, Union
from urllib.parse import urljoin
from urllib.request import getproxies

import requests
import requests.adapters
from requests.adapters import HTTPAdapter, Retry
from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor

from ansys.simai.core import __version__
from ansys.simai.core.data.types import APIResponse, File, MonitorCallback
from ansys.simai.core.errors import ConfigurationError, ConnectionError
from ansys.simai.core.utils.auth import Authenticator
from ansys.simai.core.utils.configuration import ClientConfig
from ansys.simai.core.utils.files import file_path_to_obj_file
from ansys.simai.core.utils.requests import handle_response

logger = logging.getLogger(__name__)


class TruststoreAdapter(HTTPAdapter):
    def init_poolmanager(self, *a, **kw):
        if sys.version_info < (3, 10):
            raise ConfigurationError("The system CA store can only be used with python >= 3.10")

        import truststore

        ctx = truststore.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
        return super().init_poolmanager(*a, **kw, ssl_context=ctx)


class ApiClientMixin:
    """Provides the core that all mixins and the API client are built on."""

    def __init__(self, *args, config: ClientConfig):  # noqa: D107
        self._session = requests.Session()
        if config.tls_ca_bundle == "system":
            self._session.mount("https://", TruststoreAdapter())
        elif config.tls_ca_bundle == "unsecure-none":
            self._session.verify = False
        elif isinstance(config.tls_ca_bundle, os.PathLike):
            self._session.verify = str(config.tls_ca_bundle)

        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        self._session.mount("http", HTTPAdapter(max_retries=retries))

        system_proxies = getproxies()
        if config.https_proxy:
            https_proxy_str = str(config.https_proxy)
            logger.debug(f"Connecting using specified proxy: {https_proxy_str}")
            self._session.proxies = {"https": https_proxy_str}
        elif system_proxies:
            logger.debug(f"Using detected system proxies: {system_proxies}")
            self._session.proxies = system_proxies

        self._url_prefix = config.url
        self._set_user_agent()
        self._session.auth = Authenticator(config, self._session)

    def _set_user_agent(self) -> None:
        """Set the user-agent header for the session."""
        user_agent = f"PySimAI {__version__}"
        self._session.headers.update({"User-Agent": user_agent})

    def _get(self, url, *args, **kwargs) -> APIResponse:
        """_request with method set to GET."""
        return self._request("GET", url, *args, **kwargs)

    def _put(self, url, *args, **kwargs) -> APIResponse:
        """_request with method set to PUT."""
        return self._request("PUT", url, *args, **kwargs)

    def _post(self, url, *args, **kwargs) -> APIResponse:
        """_request with method set to POST."""
        return self._request("POST", url, *args, **kwargs)

    def _delete(self, url, *args, **kwargs) -> APIResponse:
        """_request with method set to DELETE."""
        return self._request("DELETE", url, *args, **kwargs)

    def _patch(self, url, *args, **kwargs) -> APIResponse:
        """_request with method set to PATCH."""
        return self._request("PATCH", url, *args, **kwargs)

    def build_full_url_for_endpoint(self, url) -> str:  # noqa: D102
        return urljoin(str(self._url_prefix), url, allow_fragments=True)

    def _request(
        self,
        method: str,
        url,
        *args,
        return_json: bool = True,
        **kwargs,
    ) -> APIResponse:
        """Wrap around :py:meth:`requests.Session.request`.

        By default, this method expects a JSON response. If you call an endpoint that does
        not return a JSON response, specify ``return_json=False``.

        Args:
            method: HTTP verb of the request.
            url: URL of the request.
            *args: Additional arguments for the request.
            return_json: Whether the expected response is a json. If ``True``, the JSON
                is returned directly. Otherwise, the response is returned.
            **kwargs: Additional kwargs for request.

        Returns:
            JSON dictionary of the response if :py:args:`return_json` is True. The raw
                :py:class:`requests.Response` otherwise.
        """
        logger.debug(f"Request {method} on {url}")
        full_url = self.build_full_url_for_endpoint(url)
        try:
            return handle_response(
                self._session.request(method, full_url, *args, **kwargs),
                return_json=return_json,
            )
        except requests.exceptions.ConnectionError as e:
            raise ConnectionError(e) from None

    def download_file(
        self,
        download_url: str,
        file: Optional[File] = None,
        monitor_callback: Optional[MonitorCallback] = None,
        request_json_body: Optional[Dict[str, Any]] = None,
        request_method: str = "GET",
    ) -> Union[None, BinaryIO]:
        """Download a file from a URL into a file or a :class:`BytesIO` object.

        Args:
            download_url: URL for getting the file.
            file: Optional binary file or path for the downloaded file.
            monitor_callback: Optional callback to monitor the progress of the download.
                For more information, see the :obj:`~ansys.simai.core.data.types.MonitorCallback`
                object.
            request_json_body: Optional JSON to include in the request.
            request_method: HTTP verb of the request.

        Raises:
            ConnectionError: If an error occurred during the download.

        Returns:
            None if a file is provided or a ``BytesIO`` object with the file's
            content otherwise.
        """
        if file is None:
            output_file = BytesIO()
            close_file = False
        elif isinstance(file, (Path, os.PathLike, str)):
            output_file = file_path_to_obj_file(file, "wb")
            close_file = True
        else:
            output_file = file
            close_file = False

        request_kwargs = {"stream": True, "return_json": False}
        if request_json_body is not None:
            request_kwargs.update({"json": request_json_body})
        response = self._request(request_method, download_url, **request_kwargs)
        if monitor_callback is not None:
            monitor_callback(int(response.headers.get("Content-Length", 0)))
        logger.info("Starting download.")
        try:
            for chunk in response.iter_content(chunk_size=1024):
                bytes_read_delta = output_file.write(chunk)
                if monitor_callback is not None:
                    monitor_callback(bytes_read_delta)
        except requests.exceptions.ConnectionError as e:
            logger.debug("Error {e} happened during download stream.")
            if close_file is True:
                output_file.close()
                os.remove(output_file)
            raise ConnectionError(e) from None

        logger.info("Download complete.")
        if close_file is True:
            output_file.close()
        elif output_file.seekable():
            output_file.seek(0)
        if file is None:
            return output_file

    def upload_file_with_presigned_post(
        self,
        file: BinaryIO,
        presigned_post: Dict[str, Any],
        monitor_callback: Optional[MonitorCallback] = None,
    ):
        upload_form = presigned_post["fields"]
        filename = getattr(file, "name", "")
        upload_form["file"] = (filename, file, "application/octet-stream")
        multipart = MultipartEncoder(upload_form)
        if monitor_callback is not None:
            # Wrap the monitor callback so that it receives only the bytes read
            # instead of the full MultipartEncoderMonitor object
            def callback(monitor):
                monitor.previous_bytes_read = getattr(monitor, "previous_bytes_read", 0)
                new_bytes = monitor.bytes_read - monitor.previous_bytes_read
                monitor.previous_bytes_read += new_bytes
                monitor_callback(new_bytes)

            multipart = MultipartEncoderMonitor(multipart, callback)
        self._post(
            presigned_post["url"],
            data=multipart.read(),
            headers={"Content-Type": multipart.content_type},
            return_json=False,
        )

    def upload_parts(
        self,
        url: str,
        file: BinaryIO,
        upload_id: str,
        part_size: int = int(100e6),
        monitor_callback: Optional[MonitorCallback] = None,
    ) -> List[Dict[str, Any]]:
        """Upload parts using the given endpoints to get presigned ``PUT`` URLs.

        Returns:
            List of parts with their IDs and HTTP ETags.
        """
        part_number = 1
        parts = []
        while True:
            part_data = file.read(part_size)
            if len(part_data) == 0:
                break
            logger.debug(f"Uploading part {part_number}, sizeof {len(part_data)} bytes")
            create_part = self._put(url, json={"part_number": part_number, "upload_id": upload_id})
            uploaded_part = self._put(create_part["url"], data=part_data, return_json=False)
            parts.append({"PartNumber": part_number, "ETag": uploaded_part.headers["ETag"]})
            part_number += 1
            if monitor_callback is not None:
                monitor_callback(len(part_data))
        return parts



# Copyright (C) 2023 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import json
import logging
from typing import Any, Dict, List, Optional

from ansys.simai.core.api.mixin import ApiClientMixin

logger = logging.getLogger(__name__)


class PostProcessingClientMixin(ApiClientMixin):
    def run_post_processing(
        self,
        prediction_id: str,
        post_processing_type: str,
        params: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        """Run a postprocessing on the given prediction.

        If the result of the requested postprocessing already exists, the postprocessing
        is not rerun.

        Args:
            prediction_id: ID of the prediction.
            post_processing_type: Type of postprocessing to run on the prediction.
            params: Additional JSON parameters as dictionary if required.

        Returns:
            JSON with the created or existing postprocessing.
        """
        return self._post(
            f"predictions/{prediction_id}/post-processings/{post_processing_type}",
            json=params,
        )

    def get_post_processing_result(self, post_processing_id: str) -> Dict[str, Any]:
        """Get the result of a postprocessing.

        Args:
            post_processing_id: ID of the postprocessing.

        Returns:
            JSON with the result of the postprocessing.
        """
        return self._get(f"post-processings/{post_processing_id}")

    def delete_post_processing(self, post_processing_id: str):
        """Delete a postprocessing.

        Args:
            post_processing_id: ID of the postprocessing.

        Raises:
            NotFoundError: If a postprocessing with this ID is not found on the server.
        """
        return self._delete(
            f"post-processings/{post_processing_id}",
            return_json=False,
        )

    def get_post_processings_in_workspace(
        self, workspace_id: str, pp_type: Optional[str]
    ) -> List[Dict[str, Any]]:
        """Get all postprocessings in the given workspace.

        Args:
            workspace_id: ID of the target workspace.
            pp_type: Type of postprocessings to return. If this parameter is empty, all
            postprocessings are returned.
        """
        endpoint = f"post-processings/type/{pp_type}" if pp_type else "post-processings/"
        # The content of this endpoint can be paginated
        initial_request = self._get(endpoint, params={"workspace": workspace_id}, return_json=False)
        pagination_info = json.loads(initial_request.headers.get("X-Pagination", "{}"))
        post_processings = initial_request.json()

        for page in range(2, pagination_info.get("total_pages", 1) + 1):
            page_request = self._get(endpoint, params={"workspace": workspace_id, "page": page})
            post_processings.extend(page_request)
        return post_processings

    def get_post_processings_for_prediction(
        self,
        prediction_id: str,
        pp_type: Optional[str],
        filters: Optional[Dict[str, Any]] = None,
    ) -> List[Dict[str, Any]]:
        """Get all postprocessings belonging to the given prediction.

        Args:
            prediction_id: ID of the target prediction.
            pp_type: Type of postprocessings to return. If this parameter is empty, all
                postprocessings are returned.
            filters: Filters to apply to the query, if any.
        """
        endpoint = f"predictions/{prediction_id}/post-processings/"
        if pp_type:
            endpoint = endpoint + pp_type
        params = {"filters": json.dumps(filters)} if filters else None
        return self._get(endpoint, params=params)

    def post_processings_export_url(self):
        return self.build_full_url_for_endpoint("post-processings/export")



# Copyright (C) 2023 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pathlib
from typing import TYPE_CHECKING, Dict, List, Optional, Union

from ansys.simai.core.data.base import ComputableDataModel, Directory
from ansys.simai.core.data.types import (
    Filters,
    Identifiable,
    MonitorCallback,
    NamedFile,
    Path,
    SizedIterator,
    SubsetEnum,
    get_id_from_identifiable,
    to_raw_filters,
    unpack_named_file,
)
from ansys.simai.core.errors import InvalidArguments
from ansys.simai.core.utils.pagination import DataModelIterator

if TYPE_CHECKING:
    from ansys.simai.core.data.projects import Project
    from ansys.simai.core.data.training_data_parts import TrainingDataPart


def _upload_training_data_part(id, named_part, client, monitor_callback):
    with unpack_named_file(named_part) as (file, name, extension):
        (training_data_part_fields, upload_id) = client._api.create_training_data_part(
            id, name, extension
        )
        training_data_part = client.training_data_parts._model_from(
            training_data_part_fields, is_upload_complete=False
        )
        parts = client._api.upload_parts(
            f"training-data-parts/{training_data_part.id}/part",
            file,
            upload_id,
            monitor_callback=monitor_callback,
        )
        client._api.complete_training_data_part_upload(training_data_part.id, upload_id, parts)
    return training_data_part


class TrainingData(ComputableDataModel):
    """Provides the local representation of a training data object."""

    def __repr__(self) -> str:
        return f"<TrainingData: {self.id}, {self.name}>"

    @property
    def name(self) -> str:
        """Name of the training data."""
        return self.fields["name"]

    @property
    def parts(self) -> List["TrainingDataPart"]:
        """List of all :class:`parts<ansys.simai.core.data.training_data_parts.TrainingDataPart>`
        objects in the training data.
        """
        return [
            self._client.training_data_parts._model_from(training_data_part)
            for training_data_part in self.fields["parts"]
        ]

    def get_subset(self, project: Identifiable["Project"]) -> Optional[SubsetEnum]:
        """Get the subset that the training data belongs to, in relation to the given project.

        Args:
            project: ID or :class:`model <.projects.Project>` of the project to check
                the :class:`~.projects.Project` object for, or its ID.

        Returns:
            The :obj:`~ansys.simai.core.data.types.SubsetEnum` of the subset to which
            the :class:`TrainingData` belongs to if any, ``None`` otherwise.
            (e.g. <SubsetEnum.TEST: 'Test'>)
        """
        project_id = get_id_from_identifiable(project, default=self._client._current_project)
        subset_value = self._client._api.get_training_data_subset(project_id, self.id).get("subset")
        return SubsetEnum(subset_value) if subset_value else None

    def assign_subset(
        self, project: Identifiable["Project"], subset: Optional[Union[SubsetEnum, str]]
    ) -> None:
        """Assign the training data to a subset in relation to a given project.

        Args:
            project: ID or :class:`model <.projects.Project>`
            subset: SubsetEnum attribute (e.g. SubsetEnum.TRAINING) or string value (e.g. "Training") or None to unassign.
                Available options: (Training, Test)

                Each new training data added to the project will be set to "None" by default.

                None allows for resetting the subset assignment of your training data, which will be automatically allocated
                in either test or training subsets upon each model building request. As a rule of thumb, 10% of all data should
                be allocated to the test subset.

                When wanting to assign a specific subset to your training data, note that:

                Each subset requires at least one data point.
                The training subset is used to train the model. The test subset is used for the model
                evaluation report but is not learned by the model.
                It is recommended to allocate about 10% of your data to the test subset.

        Returns:
            None
        """
        if subset is not None and subset not in SubsetEnum.__members__.values():
            raise InvalidArguments("Must be None or one of: 'Training', 'Test'.")

        project_id = get_id_from_identifiable(project, default=self._client._current_project)
        self._client._api.put_training_data_subset(project_id, self.id, subset)

    @property
    def extracted_metadata(self) -> Optional[Dict]:
        """Metadata extracted from the training data."""
        return self.fields["extracted_metadata"]

    def extract_data(self) -> None:
        """Extract or reextract the data from a training data.

        Data should be extracted from a training data once all its parts have been fully uploaded.
        This is done automatically when using :meth:`TrainingDataDirectory.upload_folder` to create training data.

        Data can only be reextracted from a training data if the extraction previously failed or if new files have been added.
        """
        self._client._api.compute_training_data(self.id)

    def delete(self) -> None:
        """Delete the training data on the server."""
        self._client._api.delete_training_data(self.id)

    def upload_part(
        self, file: NamedFile, monitor_callback: Optional[MonitorCallback] = None
    ) -> "TrainingDataPart":
        """Add a part to the training data.

        Args:
            file: :obj:`~ansys.simai.core.data.types.NamedFile` to upload.
            monitor_callback: Optional callback for monitoring the progress of the download.
                For more information, see the :obj:`~ansys.simai.core.data.types.MonitorCallback`
                object.

        Returns:
            Created :class:`~ansys.simai.core.data.training_data_parts.TrainingDataPart`.
        """
        return _upload_training_data_part(self.id, file, self._client, monitor_callback)

    def upload_folder(
        self,
        folder_path: Path,
    ) -> List["TrainingDataPart"]:
        """Upload all the parts contained in a folder to a :class:`~ansys.simai.core.data.training_data.TrainingData` instance.

        Upon upload completion, SimAI will extract data from each part.

        Args:
            folder_path: Path to the folder with the files to upload.

        Returns:
            List of uploaded training data parts.
        """
        return self._directory.upload_folder(self.id, folder_path)

    def add_to_project(self, project: Identifiable["Project"]):
        """Add the training data to a :class:`~ansys.simai.core.data.projects.Project` object.

        Args:
            project: ID or :class:`model <.projects.Project>` object of the project to add the data to.
        """
        project_id = get_id_from_identifiable(project)
        self._client._api.add_training_data_to_project(self.id, project_id)

    def remove_from_project(self, project: Identifiable["Project"]):
        """Remove the training data from a :class:`~ansys.simai.core.data.projects.Project` object.

        Args:
            project: ID or :class:`model <.projects.Project>` of the project to remove data from.

        Raises:
            ansys.simai.core.errors.ApiClientError: If the data is the project's sample.
            ansys.simai.core.errors.ApiClientError: If the project is in training.
        """
        project_id = get_id_from_identifiable(project)
        self._client._api.remove_training_data_from_project(self.id, project_id)


class TrainingDataDirectory(Directory[TrainingData]):
    """Provides a collection of methods related to training data.

    This class is accessed through ``client.training_data``.

    Example:
        List all of the training data::

            import ansys.simai.core

            simai = ansys.simai.core.from_config()
            simai.training_data.list()
    """

    _data_model = TrainingData

    def iter(self, filters: Optional[Filters] = None) -> SizedIterator[TrainingData]:
        """Iterate over all :class:`TrainingData` objects on the server.

        Args:
            filters: Optional :obj:`~.types.Filters` to apply.

        Returns:
            Iterator over all :class:`TrainingData` objects on the server.
        """
        raw_filters = to_raw_filters(filters)
        raw_iterable = self._client._api.iter_training_data(filters=raw_filters)
        return DataModelIterator(raw_iterable, self)

    def list(self, filters: Optional[Filters] = None) -> List[TrainingData]:
        """List all :class:`TrainingData` objects on the server.

        Warning:
            This can take a very long time, consider using :py:meth:`~iter` instead.

        Args:
            filters: Optional :obj:`~.types.Filters` to apply.

        Returns:
            List of all :class:`TrainingData` objects on the server.
        """
        return list(self.iter(filters))

    def get(self, id) -> TrainingData:
        """Get a specific :class:`TrainingData` object from the server."""
        return self._model_from(self._client._api.get_training_data(id))

    def delete(self, training_data: Identifiable[TrainingData]) -> None:
        """Delete a :class:`TrainingData` object and its associated parts from the server.

        Args:
            training_data: ID or :class:`model <TrainingData>` object of the :class:`TrainingData` object.
        """
        return self._client._api.delete_training_data(get_id_from_identifiable(training_data))

    def create(self, name: str, project: Optional[Identifiable["Project"]] = None) -> TrainingData:
        """Create a :class:`TrainingData` object.

        Args:
            name: Name to give the new :class:`TrainingData` object.
            project: :class:`~.projects.Project` object to associate the data with.

        Returns:
            Created :class:`TrainingData` object.
        """
        project_id = get_id_from_identifiable(project, required=False)
        return self._model_from(self._client._api.create_training_data(name, project_id))

    def upload_part(
        self,
        training_data: Identifiable[TrainingData],
        file: NamedFile,
        monitor_callback: Optional[MonitorCallback],
    ) -> "TrainingDataPart":
        """Add a part to a :class:`TrainingData` object.

        Args:
            training_data: ID or :class:`model <TrainingData>` object of the training data to
                add the part to.
            file: :obj:`~ansys.simai.core.data.types.NamedFile` to upload.
            monitor_callback: Optional callback for monitoring the progress of the upload.
                For more information, see the :obj:`~ansys.simai.core.data.types.MonitorCallback`
                object.

        Returns:
            Added :class:`~ansys.simai.core.data.training_data_parts.TrainingDataPart` object.
        """
        return _upload_training_data_part(
            get_id_from_identifiable(training_data),
            file,
            self._client,
            monitor_callback,
        )

    def upload_folder(
        self,
        training_data: Identifiable[TrainingData],
        folder_path: Path,
    ) -> List["TrainingDataPart"]:
        """Upload all files in a folder to a :class:`~ansys.simai.core.data.training_data.TrainingData` object.

        This method automatically requests computation of the training data once the upload is complete
        unless specified otherwise.

        Args:
            training_data: ID or :class:`model <TrainingData>` object of the training data to upload parts to.
            folder_path: Path to the folder that contains the files to upload.
        """
        training_data_id = get_id_from_identifiable(training_data)
        path = pathlib.Path(folder_path)
        if not path.is_dir():
            raise InvalidArguments("Provided path is not a folder.")
        path_content = path.glob("[!.]*")
        files = (obj for obj in path_content if obj.is_file())
        uploaded_parts = []
        for file in files:
            uploaded_parts.append(
                _upload_training_data_part(training_data_id, file, self._client, None)
            )
        self._client._api.compute_training_data(training_data_id)
        return uploaded_parts



# Copyright (C) 2023 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
from typing import Any, Dict, List, Optional

from ansys.simai.core.data.base import ComputableDataModel, Directory
from ansys.simai.core.data.geometries import Geometry
from ansys.simai.core.data.post_processings import PredictionPostProcessings
from ansys.simai.core.data.types import (
    BoundaryConditions,
    Identifiable,
    build_boundary_conditions,
    get_id_from_identifiable,
)
from ansys.simai.core.data.workspaces import Workspace

logger = logging.getLogger(__name__)


class Prediction(ComputableDataModel):
    """Provides the local representation of a prediction object."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._post_processings = PredictionPostProcessings(self)
        self._geometry = None

    @property
    def geometry_id(self) -> str:
        """ID of the parent geometry.

        See Also:
            - :attr:`geometry`: Get the parent geometry.
        """
        return self.fields["geometry_id"]

    @property
    def geometry(self) -> Geometry:
        """Parent geometry.

        The parent geometry is queried if is not already known by the current
        SimAI client session.

        See Also:
            - :attr:`geometry_id`: Get the parent geometry's ID without query.
        """
        if self._geometry is None:
            if self.geometry_id in self._client.geometries._registry:
                self._geometry = self._client.geometries._registry[self.geometry_id]
            else:
                self._geometry = self._client.geometries.get(id=self.geometry_id)
        return self._geometry

    @property
    def boundary_conditions(self) -> BoundaryConditions:
        """Boundary conditions of the prediction."""
        return self.fields["boundary_conditions"]

    @property
    def post(self) -> PredictionPostProcessings:
        """Namespace containing methods for postprocessing the result of a prediction.

        For more information, see the :py:class:`~ansys.simai.core.data.post_processings.PredictionPostProcessings`
        class.
        """
        return self._post_processings

    @property
    def confidence_score(self) -> Optional[str]:
        """Confidence score, which is either ``high`` or ``low``.

        This method blocks until the confidence score is computed.
        """
        self.wait()
        confidence_score = self.fields["confidence_score"]
        if confidence_score not in ["high", "low", None]:
            raise ValueError("Must be None or one of: 'high', 'low', None.")
        return confidence_score

    @property
    def raw_confidence_score(self) -> Optional[float]:
        """Raw confidence score, a float.

        This method blocks until the confidence score is computed.
        """
        self.wait()
        return self.fields["raw_confidence_score"]

    def delete(self) -> None:
        """Remove a prediction from the server."""
        self._client._api.delete_prediction(self.id)
        self._unregister()

    def _wait_all(self):
        """Wait until both this prediction and any postprocessing launched on it
        have finished processing.

        This method blocks until both the prediction and any postprocessing launched
        locally have either finished processing or have failed.

        Postprocessing launched by other SimAI client sessions or on the front-end
        are not waited upon.
        """
        # wait for own creation Event
        logger.debug("prediction: waiting for own loading")
        super().wait()
        if self.has_failed:
            return
        # Wait for its post-processings if any
        if self.post._local_post_processings:
            logger.debug("prediction: waiting for postprocessings loading")
            for post_processing in self.post._local_post_processings:
                post_processing.wait()

    def _merge_fields_from_results(self, results: dict):
        super()._merge_fields_from_results(results)

        # confidence_score is in result/data/values
        values = results.get("data", {}).get("values", {})
        if "confidence_score" in values:
            self.fields["confidence_score"] = values["confidence_score"]


class PredictionDirectory(Directory[Prediction]):
    """Provides a collection of methods related to model predictions.

    This method is accessed through ``client.prediction``.

    Example:
        .. code-block:: python

            import ansys.simai.core

            simai = ansys.simai.core.from_config()
            simai.predictions.list()
    """

    _data_model = Prediction

    @property
    def boundary_conditions(self) -> Dict[str, Any]:
        """Information on the boundary conditions expected by the model of the current workspace.
        For example, the prediction's input.
        """
        return self._client.current_workspace.model_manifest.boundary_conditions

    @property
    def physical_quantities(self) -> Dict[str, Any]:
        """Information on the physical quantities generated by the model. For example, the
        prediction's output.
        """
        return self._client.current_workspace.model_manifest.physical_quantities

    @property
    def info(self):
        """Information on the prediction's inputs and outputs.

        Example:
            .. code-block:: python

                from pprint import pprint
                import ansys.simai.core

                simai = ansys.simai.core.from_config()
                prediction_info = simai.predictions.info
                pprint(prediction_info)
        """
        return {
            "boundary_conditions": self.boundary_conditions,
            "physical_quantities": self.physical_quantities,
        }

    def list(self, workspace: Optional[Identifiable[Workspace]] = None) -> List[Prediction]:
        """List all predictions on the server that belong to the specified workspace or the configured one.

        Args:
            workspace: ID or :class:`model <.workspaces.Workspace>` of the workspace to list the predictions for.
                This parameter is necessary if no workspace is set for the client.
        """
        workspace_id = get_id_from_identifiable(workspace, default=self._client._current_workspace)
        raw_predictions = self._client._api.predictions(workspace_id)
        return list(map(self._model_from, raw_predictions))

    def get(self, id: str) -> Prediction:
        """Get a specific prediction object from the server by ID.

        Args:
            id: ID of the prediction.

        Returns:
            :class:`Prediction` instance with the given ID if it exists.

        Raises:
            :class:`NotFoundError`: No prediction with the given ID exists.
        """
        return self._model_from(self._client._api.get_prediction(id))

    def delete(self, prediction: Identifiable[Prediction]) -> None:
        """Delete a specific prediction from the server.

        Args:
            prediction: ID or :class:`model <Prediction>` of the prediction.

        Raises:
            :py:class:`ansys.simai.core.errors.NotFoundError`: No prediction with the given ID exists.
        """
        prediction_id = get_id_from_identifiable(prediction)
        self._client._api.delete_prediction(prediction_id)
        self._unregister_item_with_id(prediction_id)

    def run(  # noqa: D417
        self,
        geometry: Identifiable[Geometry],
        boundary_conditions: Optional[BoundaryConditions] = None,
        **kwargs,
    ) -> Prediction:
        """Run a prediction on a given geometry with a given boundary conditions.

        Boundary conditions can be passed as a dictionary or as kwargs.

        To learn more about the expected boundary conditions in your workspace, you can use the
        ``simai.current_workspace.model_manifest.boundary_conditions`` or ``simai.predictions.boundary_conditions``
        method, where ``ex`` is your `~ansys.simai.core.client.SimAIClient` object.

        Args:
            geometry: ID or :class:`model <.geometries.Geometry>` of the target geometry.
            boundary_conditions: Boundary conditions to apply in dictionary form.

        Returns:
            Created prediction object.

        Raises:
            ProcessingError: If the server failed to process the request.

        Examples:
            .. code-block:: python

                simai = ansys.simai.core.from_config()
                geometry = simai.geometries.list()[0]
                prediction = simai.predictions.run(geometry, dict(Vx=10.5, Vy=2))

            Using kwargs:

            .. code-block:: python

                prediction = simai.predictions.run(geometry_id, Vx=10.5, Vy=2)
        """
        bc = build_boundary_conditions(boundary_conditions, **kwargs)
        geometry = self._client.geometries.get(id=get_id_from_identifiable(geometry))
        prediction = geometry.run_prediction(boundary_conditions=bc)
        for location, warning_message in prediction.fields.get("warnings", {}).items():
            logger.warning(f"{location}: {warning_message}")
        return prediction



# Copyright (C) 2023 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Dict, List, Optional, Union

from ansys.simai.core.data.geometries import Geometry
from ansys.simai.core.data.predictions import Prediction
from ansys.simai.core.data.selection_post_processings import SelectionPostProcessingsMethods
from ansys.simai.core.data.types import (
    BoundaryConditions,
    are_boundary_conditions_equal,
    is_boundary_conditions,
)
from ansys.simai.core.errors import _foreach_despite_errors
from ansys.simai.core.utils.numerical import (
    DEFAULT_COMPARISON_EPSILON,
    validate_tolerance_parameter,
)
from ansys.simai.core.utils.validation import _enforce_as_list_passing_predicate


class Point:
    """Provides a ``Point`` object, where a prediction can be run.

    A point is at the intersection of a :class:`~ansys.simai.core.data.geometries.Geometry`
    isstance and :class:`~ansys.simai.core.data.types.BoundaryConditions` instance.
    """

    def __init__(self, geometry: Geometry, boundary_conditions: BoundaryConditions):
        self._geometry = geometry
        self._boundary_conditions = boundary_conditions
        self._prediction: Optional[Prediction] = None

    @property
    def geometry(self) -> Geometry:
        """:class:`~ansys.simai.core.data.geometries.Geometry` object for the :class:`Point` instance."""
        return self._geometry

    @property
    def boundary_conditions(self) -> BoundaryConditions:
        """:class:`~ansys.simai.core.data.types.BoundaryConditions` object for the :class:`Point`
        instance.
        """
        return self._boundary_conditions

    @property
    def prediction(self) -> Union[Prediction, None]:
        """:class:`~ansys.simai.core.data.predictions.Prediction` instance
        corresponding to the point or ``None`` if no prediction has yet been run.
        """
        return self._prediction

    def run_prediction(self, boundary_conditions: BoundaryConditions):
        """Run the prediction on the geometry for this boundary condition."""
        self._prediction = self._geometry.run_prediction(boundary_conditions=boundary_conditions)

    def __repr__(self):
        return "{}(geometry: {}, boundary_condition: {}, prediction: {})".format(
            self.__class__, self._geometry, self._boundary_conditions, self.prediction
        )


class Selection:
    """Provides a ``Selection`` object, which is a collection of :class:`Points <Point>` instances.

    Selections are built from a list of :class:`Geometries <ansys.simai.core.data.geometries.Geometry>`
    instances and a list of :class:`~ansys.simai.core.data.types.BoundaryConditions` instances.

    The resulting selection contains all combinations between the geometries
    and the boundary conditions.

    Args:
        geometries: Geometries to include in the selection.
        boundary_conditions: Boundary conditions to include in the selection.
        tolerance: Optional delta to apply to boundary condition equality.
                The default is ``10**-6``. If the difference between two boundary
                conditions is lower than the tolerance, the two boundary conditions
                are considered as equal.
    """

    def __init__(
        self,
        geometries: Union[Geometry, List[Geometry]],
        boundary_conditions: Union[BoundaryConditions, List[BoundaryConditions]],
        tolerance: Optional[float] = None,
    ):
        # Validate parameters
        geometries = _enforce_as_list_passing_predicate(
            geometries,
            lambda g: isinstance(g, Geometry),
            "'geometries' must be a geometry or a list of 'Geometry' objects.",
        )
        boundary_conditions = _enforce_as_list_passing_predicate(
            boundary_conditions,
            lambda bc: is_boundary_conditions(bc),
            "'boundary_conditions' must be a dictionary of numbers.",
        )
        if tolerance is None:
            tolerance = DEFAULT_COMPARISON_EPSILON
        validate_tolerance_parameter(tolerance)
        self.tolerance = tolerance

        points = []
        for geometry in geometries:
            for boundary_condition in boundary_conditions:
                points.append(Point(geometry, boundary_condition))
        self._geometries = geometries
        self._boundary_conditions = boundary_conditions
        self._points = points
        self._post_processings = SelectionPostProcessingsMethods(self)
        self.reload()

    @property
    def points(self) -> List[Point]:
        """List of all :class:`Points <Point>` instances in the selection."""
        return self._points

    @property
    def predictions(self) -> List[Prediction]:
        """List of all existing :class:`Prediction <ansys.simai.core.data.predictions.Prediction>`
        instances in the selection.
        """
        return self.get_predictions()

    @property
    def geometries(self) -> List[Geometry]:
        """List of all existing :class:`Geometries <ansys.simai.core.data.geometries.Geometry>`
        instances in the selection.
        """
        return self._geometries

    @property
    def boundary_conditions(self) -> List[BoundaryConditions]:
        """List of all existing :class:`BoundaryConditions <ansys.simai.core.data.types.BoundaryConditions>`
        instances in the selection.
        """
        return self._boundary_conditions

    @property
    def points_with_prediction(self) -> List[Optional[Point]]:
        """List of all :class:`Points <Point>` instances in the selection where predictions exist."""
        return [(point if point.prediction else None) for point in self.points]

    @property
    def points_without_prediction(self) -> List[Optional[Point]]:
        """List of all :class:`Points <Point>` instances in the selection where predictions don't exist."""
        return [(point if point.prediction is None else None) for point in self.points]

    def get_predictions(self) -> List[Prediction]:  # noqa D102
        return [point.prediction for point in self.points if point.prediction is not None]

    def get_runnable_predictions(self) -> List[Point]:
        """List of all :class:`Points <Point>` instances in the selection where predictions haven't
        been run yet.
        """
        return [point for point in self.points if point.prediction is None]

    def run_predictions(self) -> None:
        """Run all missing predictions in the selection."""
        _foreach_despite_errors(
            lambda point: point.run_prediction(boundary_conditions=point.boundary_conditions),
            self.get_runnable_predictions(),
        )

    def wait(self) -> None:
        """Wait for all ongoing operations (predictions and postprocessings)
        in the selection to finish.

        Raises:
            ansys.simai.core.errors.SimAIError: If a single error occurred when computing this selection's operations.
            ansys.simai.core.errors.MultipleErrors: If multiple exceptions occurred when computing this selection's operations.
        """
        _foreach_despite_errors(lambda prediction: prediction._wait_all(), self.get_predictions())

    def reload(self) -> None:
        """Refreshes the predictions in the selection.

        This method loads any predictions run from another session and
        removes possible deleted predictions.
        """
        _predictions_by_geometry_id: Dict[str, List[Prediction]] = {}
        for point in self.points:
            geometry = point.geometry
            if geometry.id not in _predictions_by_geometry_id:
                _predictions_by_geometry_id[geometry.id] = geometry.get_predictions()
            predictions = _predictions_by_geometry_id[geometry.id]
            try:
                point._prediction = next(
                    (
                        s
                        for s in predictions
                        if are_boundary_conditions_equal(
                            s.boundary_conditions,
                            point.boundary_conditions,
                            self.tolerance,
                        )
                    )
                )
            except StopIteration:
                point._prediction = None

    @property
    def post(self) -> SelectionPostProcessingsMethods:
        """Namespace containing methods to access and run postprocessings
        for the predictions in the selection.

        For more information, see the :py:class:`~ansys.simai.core.data.selection_post_processings.SelectionPostProcessingsMethods`
        class.

        """
        return self._post_processings



# Copyright (C) 2023 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import TYPE_CHECKING, Any, Callable, TypeVar, cast

if TYPE_CHECKING:
    from typing_extensions import Concatenate, ParamSpec

    P = ParamSpec("P")
    T = TypeVar("T")


def steal_kwargs_type_on_method(
    original_fn: "Callable[P, Any]",
) -> "Callable[[Callable], Callable[Concatenate[Any, P], T]]":
    """Takes type hints from ``original_fn`` and apply them to the decorated method.
    The ``self`` argument of the method is typed ``Any``.
    """

    def return_func(func: "Callable[..., T]") -> "Callable[Concatenate[Any, P], T]":
        return cast("Callable[Concatenate[Any, P], T]", func)

    return return_func



# Copyright (C) 2023 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import warnings
from pprint import pformat
from typing import Any, BinaryIO, Dict, List, Optional, Union

from ansys.simai.core.data.base import DataModel, Directory
from ansys.simai.core.data.types import File, Identifiable, get_id_from_identifiable


class ModelManifest:
    """Provides information about a model associated with a workspace."""

    def __init__(self, raw_manifest: Dict[str, Any]):
        self._raw = raw_manifest

    def __repr__(self) -> str:
        return pformat(self._raw)

    @property
    def name(self) -> str:
        """Name of the model."""
        return self._raw["model_name"]

    @property
    def description(self) -> str:
        """Short description of the model."""
        return self._raw["description"]

    @property
    def geometry(self) -> Dict[str, Any]:
        """Information on the geometry format expected by the model."""
        return self._raw["geometry"]

    @property
    def boundary_conditions(self) -> Dict[str, Any]:
        """Information on the boundary conditions expected by the model. For example, the prediction's input."""
        return self._raw["boundary_conditions"]

    @property
    def physical_quantities(self) -> Dict[str, Any]:
        """Information on the physical quantities generated by the model. For example, the prediction's output."""
        return self._raw["physical_quantities"]

    @property
    def post_processings(self) -> "List[Dict[str, Any]]":
        """Information on the postprocessings available for the model
        and the accepted parameters when relevant.
        """
        return self._raw["available-post-processings"]


class Workspace(DataModel):
    """Provides the local representation of a workspace object."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._model_manifest = None

    def __repr__(self) -> str:
        return f"<Workspace: {self.id}, {self.name}>"

    @property
    def name(self) -> str:
        """Name of the workspace."""
        return self.fields["name"]

    @property
    def model(self) -> ModelManifest:
        """Deprecated alias to :py:attr:`~model_manifest`."""
        warnings.warn(
            "workspace.model is deprecated, please use workspace.model_manifest", stacklevel=2
        )
        return self.model_manifest

    @property
    def model_manifest(self) -> ModelManifest:
        """:class:`~ansys.simai.core.data.workspaces.ModelManifest` instance containing
        information about the model associated with the workspace.
        """
        if self._model_manifest is None:
            self._model_manifest = ModelManifest(
                self._client._api.get_workspace_model_manifest(self.id)
            )
        return self._model_manifest

    def delete(self):
        """Delete the workspace."""
        return self._client._api.delete_workspace(self.id)

    def set_as_current_workspace(self) -> None:
        """Configure the client to use this workspace instead of the one currently configured."""
        self._client.current_workspace = self

    def download_model_evaluation_report(
        self, file: Optional[File] = None
    ) -> Union[None, BinaryIO]:
        """Download the PDF of the model evaluation report for the workspace.

        Args:
            file: Binary file-object or the path of the file to put the content into.

        Returns:
            ``None`` if a file is specified or a binary file-object otherwise.
        """
        return self._client._api.download_workspace_model_evaluation_report(self.id, file)

    def download_mer_data(self, file: Optional[File] = None) -> Union[None, BinaryIO]:
        """Download the names, subsets and plotting data of the files used in the AI model and MER for the workspace.

        Args:
            file: Binary file-object or the path of the file to put the content into.

        Returns:
            ``None`` if a file is specified or a binary file-object otherwise.
        """
        return self._client._api.download_workspace_mer_data(self.id, file)


class WorkspaceDirectory(Directory[Workspace]):
    """Provides a collection of methods related to workspaces.

    This class is accessed through ``client._workspaces``.

    Example:
      .. code-block:: python

            import ansys.simai.core

            simai = ansys.simai.core.from_config()
            simai.workspaces.list()
    """

    _data_model = Workspace

    def list(self) -> List[Workspace]:
        """List all workspaces from the server."""
        return [self._model_from(workspace) for workspace in self._client._api.workspaces()]

    def get(self, id: Optional[str] = None, name: Optional[str] = None) -> Workspace:
        """Get a specific workspace object from the server by either ID or name.

        You can specify either the ID or the name, not both.

        Args:
            id: ID of the workspace.
            name: Name of the workspace.

        Returns:
            Workspace.

        Raises:
            NotFoundError: No geometry with the given ID exists.
        """
        if name and id:
            raise ValueError("'id' and 'name' cannot both be specified.")
        if name:
            return self._model_from(self._client._api.get_workspace_by_name(name))
        if id:
            return self._model_from(self._client._api.get_workspace(id))
        raise ValueError("Either 'id' or 'name' must be specified.")

    def create(self, name: str, model_id: str) -> Workspace:
        """Create a workspace.

        Args:
            name: Name to give the new workspace.
            model_id: ID of the model for the workspace to use.
        """
        return self._model_from(self._client._api.create_workspace(name, model_id))

    def delete(self, workspace: Identifiable[Workspace]) -> None:
        """Delete a workspace.

        Args:
            workspace: ID or :class:`model <Workspace>` of the workspace.
        """
        self._client._api.delete_workspace(get_id_from_identifiable(workspace))



# Copyright (C) 2023 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
from typing import TYPE_CHECKING, Dict, List, NamedTuple, Optional

from ansys.simai.core.data.base import DataModel, Directory
from ansys.simai.core.data.model_configuration import ModelConfiguration
from ansys.simai.core.data.types import Identifiable, get_id_from_identifiable
from ansys.simai.core.errors import InvalidArguments, ProcessingError
from ansys.simai.core.utils.numerical import cast_values_to_float

if TYPE_CHECKING:
    from ansys.simai.core.data.global_coefficients_requests import (
        CheckGlobalCoefficient,
        ComputeGlobalCoefficient,
    )
    from ansys.simai.core.data.training_data import TrainingData

EXTRA_CALCULETTE_FIELDS = ["Area", "Normals", "Centroids"]

logger = logging.getLogger(__name__)


class IsTrainableInfo(NamedTuple):
    """Properties for project's trainability.

    The objects of this class can be used as booleans
    in condition statements as in the example:

    Example:
        Verify the project is trainable

        .. code-block:: python

            pt = my_project.is_trainable()

            if pt:
                print(pt)

        It prints:

        .. code-block:: shell

            <is_trainable: False, reason(s): Not enough data to train a model: we need at least 3 data points to train a model.>

    Attributes:
        is_trainable (bool):    True if the project is trainable, False if it is not.
        reason (str):           If not_trainable is False, the reason why the project is not trainable. None otherwise.

    """

    is_trainable: bool
    reason: str = None

    def __bool__(self) -> bool:
        return self.is_trainable

    def __repr__(self) -> str:
        return f"<is_trainable: {self.is_trainable}, reason(s): {self.reason}>"


class Project(DataModel):
    """Provides the local representation of a  project object."""

    def __repr__(self) -> str:
        return f"<Project: {self.id}, {self.name}>"

    @property
    def name(self) -> str:
        """Name of project."""
        return self.fields["name"]

    @name.setter
    def name(self, new_name: str):
        """Rename the project.

        Args:
            new_name: New name to give to the project.
        """
        self._client._api.update_project(self.id, name=new_name)
        self.reload()

    @property
    def data(self) -> List["TrainingData"]:
        """List of all :class:`~ansys.simai.core.data.training_data.TrainingData` instances in the project."""
        raw_td_list = self._client._api.iter_training_data_in_project(self.id)
        return [
            self._client.training_data._model_from(training_data) for training_data in raw_td_list
        ]

    @property
    def sample(self) -> Optional["TrainingData"]:
        """Sample of the project. The sample determines what variable and settings are available during model configuration."""
        raw_sample = self.fields["sample"]
        if raw_sample is None:
            return None
        return self._client.training_data._model_from(self.fields["sample"])

    @sample.setter
    def sample(self, new_sample: Identifiable["TrainingData"]):
        td_id = get_id_from_identifiable(new_sample)
        self._client._api.set_project_sample(self.id, td_id)
        self.reload()

    @property
    def last_model_configuration(self) -> ModelConfiguration:
        """Last :class:`configuration <ansys.simai.core.data.model_configuration.ModelConfiguration>` used for model training in this project."""
        return ModelConfiguration._from_payload(
            project=self, **self.fields.get("last_model_configuration")
        )

    def is_trainable(self) -> IsTrainableInfo:
        """Check if the project meets the prerequisites to be trained."""
        tt = self._client._api.is_project_trainable(self.id)
        return IsTrainableInfo(**tt)

    def delete(self) -> None:
        """Delete the project."""
        self._client._api.delete_project(self.id)

    def get_variables(self) -> Optional[Dict[str, List[str]]]:
        """Get the available variables for the model's input/output."""
        if not self.sample:
            return None

        sample_metadata = self.sample.fields.get("extracted_metadata")
        data = {}
        for key, vals in sample_metadata.items():
            local_fields = vals.get("fields", [])
            data[key] = [local_field.get("name") for local_field in local_fields]
        return data

    def verify_gc_formula(
        self, gc_formula: str, bc: list[str] = None, surface_variables: list[str] = None
    ):
        """Verify whether the syntax of the global coefficient formula is valid."""

        if not self.sample:
            raise ProcessingError(
                f"No sample is set in the project {self.id}. A sample should be set to verify a global coefficient formula."
            )

        sample_metadata = self.sample.fields.get("extracted_metadata")

        gc_check: CheckGlobalCoefficient = self._client._check_gc_formula_directory._model_from(
            data={
                "id": f"{self.id}-check-{gc_formula}",
            },
            project_id=self.id,
            gc_formula=gc_formula,
            sample_metadata=sample_metadata,
            bc=bc,
            surface_variables=surface_variables,
        )

        gc_check.run()
        gc_check.wait()

        return gc_check.is_ready

    def compute_gc_formula(
        self, gc_formula: str, bc: list[str] = None, surface_variables: list[str] = None
    ) -> float:
        """Compute the result of a global coefficient formula according to the project sample."""

        if not self.sample:
            raise ProcessingError(
                f"No sample is set for the project {self.id}. A sample should be set for computing the results of a Global Coefficients formula."
            )

        sample_metadata = self.sample.fields.get("extracted_metadata")

        gc_compute: ComputeGlobalCoefficient = (
            self._client._compute_gc_formula_directory._model_from(
                data={
                    "id": f"{self.id}-compute-{gc_formula}",
                },
                project_id=self.id,
                gc_formula=gc_formula,
                sample_metadata=sample_metadata,
                bc=bc,
                surface_variables=surface_variables,
            )
        )

        gc_compute.run()
        gc_compute.wait()

        return cast_values_to_float(gc_compute.result["value"])

    def cancel_build(self):
        """Cancels a build if there is one pending."""

        self.reload()
        if self.fields.get("is_being_trained") is False:
            raise ProcessingError("No build pending for this project.")
        self._client._api.cancel_build(self.id)


class ProjectDirectory(Directory[Project]):
    """Provides a collection of methods related to projects.

    This class is accessed through ``client.projects``.

    Example:
        List all projects::

            import ansys.simai.core

            simai = ansys.simai.core.from_config()
            simai.projects.list()
    """

    _data_model = Project

    def list(self) -> List[Project]:
        """List all projects available on the server."""
        return [self._model_from(data) for data in self._client._api.projects()]

    def create(self, name: str) -> Project:
        """Create a project."""
        return self._model_from(self._client._api.create_project(name=name))

    def get(self, id: Optional[str] = None, name: Optional[str] = None):
        """Get a project by either ID or name.

        You can specify either the ID or the name, not both.

        Args:
            id: ID of the project.
            name: Name of the project.
        """
        if name and id:
            raise InvalidArguments("Only the 'id' or 'name' argument should be specified.")
        elif name:
            return self._model_from(self._client._api.get_project_by_name(name))
        elif id:
            return self._model_from(self._client._api.get_project(id))
        else:
            raise InvalidArguments("Either the 'id' or 'name' argument should be specified.")

    def delete(self, project: Identifiable[Project]) -> None:
        """Delete a project.

        Args:
            project: ID or :class:`model <Project>` of the project.
        """
        self._client._api.delete_project(get_id_from_identifiable(project))

    def cancel_build(self, project: Identifiable[Project]):
        """Cancel a build if one is in progress.

        Args:
            project: ID or :class:`model <Project>` of the project.
        """

        project_id = get_id_from_identifiable(project)
        project_response = self._client._api.get_project(project_id)
        if project_response.get("is_being_trained") is False:
            raise ProcessingError("No build pending for this project.")
        self._client._api.cancel_build(project_id)



# Copyright (C) 2023 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import json
import logging
import numbers
import sys
from abc import ABC, abstractmethod
from inspect import cleandoc
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Type, Union

from ansys.simai.core.data.base import ComputableDataModel, Directory
from ansys.simai.core.data.downloads import DownloadableResult
from ansys.simai.core.data.types import Identifiable, get_id_from_identifiable
from ansys.simai.core.errors import (
    InvalidArguments,
    InvalidOperationError,
    InvalidServerStateError,
)
from ansys.simai.core.utils.numerical import (
    cast_values_to_float,
    convert_axis_and_coordinate_to_plane_eq_coeffs,
)

if TYPE_CHECKING:
    from ansys.simai.core.data.predictions import Prediction
    from ansys.simai.core.data.workspaces import Workspace


logger = logging.getLogger(__name__)


class PostProcessing(ComputableDataModel, ABC):
    """Provides the local representation of a ``PostProcessing`` object.

    This is an abstract class. Depending on the postprocessing, a different implementation
    is returned. For more information, see :ref:`available_pp`.
    """

    # NOTE for developers: New postprocessings must be added to the root ``__init__.py`` file.

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._results = None
        # Set by the Directory, or else on first access
        self._prediction = None

    @property
    def parameters(self) -> Optional[Dict[str, Any]]:
        """Parameters used to run the postprocessing."""
        return self.fields["location"]

    @property
    def prediction_id(self) -> str:
        """Parent prediction's ID.

        See Also:
            - :attr:`prediction`: Get the parent prediction.
        """
        return self.fields["prediction_id"]

    @property
    def prediction(self) -> "Prediction":
        """Parent prediction.

        The parent prediction is queried if it is not already known by the current SimAPI client session.

        See Also:
            - :attr:`prediction_id`: Get the parent prediction's ID without query.
        """
        if self._prediction is None:
            if self.prediction_id in self._client.predictions._registry:
                self._prediction = self._client.predictions._registry[self.prediction_id]
            else:
                self._prediction = self._client.predictions.get(id=self.prediction_id)
        return self._prediction

    @property
    def type(self) -> str:
        """Type of postprocessing that this object represents."""
        return self._fields["type"]

    @property
    @abstractmethod
    def data(self):
        """Get the data generated by the postprocessing.

        The return type may vary depending on the postprocessing. It can be a dictionary
        or, if the data is binary, a :class:`DownloadableResult` object, which provides
        helpers to download the data into a file or in memory.
        """

    @classmethod
    def _api_name(cls) -> str:
        # Name of the postprocessing in API calls. Override if different from the class name.
        return cls.__name__

    def _get_results(self, cache=True):
        """Internal method for getting the results of this postprocessing.

        Args:
            cache:
                Whether to save results to the cache. The default is ``True``,
                which saves results to the cache. If ``False`` download
                links expire. Thus, results must be queried just before download.
        """
        if cache and self._results:
            return self._results
        res = self._client._api.get_post_processing_result(self.id)
        if cache:
            self._results = res
        return res

    def delete(self):
        """Delete the postprocessing and its result data.

        Raises:
            NotFoundError: If the postprocessing has already been deleted.
        """
        self._client._api.delete_post_processing(self.id)
        self._unregister()
        if self._prediction is not None:
            self._prediction._post_processings._delete_local_post_processing(self)


class ExportablePostProcessing(PostProcessing, ABC):
    def export(self, format: Optional[str] = "json") -> DownloadableResult:
        """Export the postprocessing results in the desired format.

        Accessing this property blocks until the data is ready.

        Args:
            format: Format to export the data in. The default is ``'json'``.
                Options are ``'csv.zip'``, ``'json'``, and ``'xlsx'``. Note that
                the ``'csv.zip'`` option exports a ZIP file containing
                multiple CSV sheets.

        Returns:
            :class:`DownloadableResult` object for downloading the exported
            data into a file or access it in memory.
        """
        self.wait()
        return DownloadableResult(
            self._client._api.post_processings_export_url(),
            self._client,
            request_method="POST",
            request_json_body={"ids": [self.id], "format": format},
        )


class GlobalCoefficients(ExportablePostProcessing):
    """Provides the representation of the global coefficients of a prediction.

    The data attribute contains a dictionary representing the global coefficients
    with its pressure and velocity components.

    This class is generated through the :meth:`PredictionPostProcessings.global_coefficients()`
    method.
    """

    @property
    def data(self) -> Dict[str, List]:
        """Dictionary containing the global coefficients, including pressure
        and velocity components.

        Accessing this property blocks until the data is ready.
        """
        self.wait()

        results = self._get_results()
        return {
            k: {**v, "data": cast_values_to_float(v["data"])}
            for k, v in results["data"]["values"].items()
        }


class SurfaceEvolution(ExportablePostProcessing):
    """Provides the representation of the ``SurfaceEvolution`` object.

    This class is generated through :meth:`PredictionPostProcessings.surface_evolution()`
    """

    @property
    def data(self) -> DownloadableResult:
        """:class:`DownloadableResult` object that allows access to the
        ``SurfaceEvolution`` JSON data, both directly in memory any by downloading it
        into a file.

        Accessing this property blocks until the data is ready.
        """
        self.wait()
        results = self._get_results(cache=False)
        return DownloadableResult(results["data"]["resources"]["json"], self._client)

    def as_dict(self) -> Dict[str, Any]:
        """Download the SurfaceEvolution JSON data and load it as a Python dictionary.

        Accessing this help method blocks until the data is ready.
        """
        return json.load(self.data.in_memory())

    @classmethod
    def _api_name(cls) -> str:
        # Name of the postprocessing in API calls. Overriding because the endpoint is SurfaceEvol not SurfaceEvolution
        return "SurfaceEvol"


class Slice(PostProcessing):
    """Provides a representation of a slice from the prediction in PNG or VTP format.

    This class is generated through the :meth:`PredictionPostProcessings.slice` method.
    """

    @property
    def data(self) -> DownloadableResult:
        """:class:`DownloadableResult` object that allows
        access to slice data, both directly in memory
        and by downloading it into a file.

        Accessing this property blocks until the data is ready.

        Returns:
            A :class:`DownloadableResult`
        """
        self.wait()
        results = self._get_results(cache=False)
        return DownloadableResult(
            results["data"]["resources"][self.parameters.get("output_format", "png")],
            self._client,
        )


class _PostProcessingVTKExport(PostProcessing, ABC):
    """Provides the representation of the result of the prediction in a format of the VTK family."""

    @property
    def data(self) -> DownloadableResult:
        """:class:`DownloadableResult` object that allows
        access to the VTK data, either directly in memory
        or by downloading it into a file.

        Accessing this property blocks until the data is ready.
        """
        self.wait()
        results = self._get_results(cache=False)
        return DownloadableResult(results["data"]["resources"]["vtk"], self._client)


class VolumeVTU(_PostProcessingVTKExport):
    """Provides for exporting the volume of the prediction in VTU format.

    This class is generated through the :meth:`PredictionPostProcessings.volume_vtu()` method.
    """


class SurfaceVTP(_PostProcessingVTKExport):
    """Exports the surface of the prediction in VTP format associating all data with cells.

    This class is generated through the :meth:`~PredictionPostProcessings.surface_vtp()` method.
    """


class SurfaceVTPTDLocation(_PostProcessingVTKExport):
    """Exports the surface of the prediction in VTP format keeping the original data association.

    This class is generated through the :meth:`~PredictionPostProcessings.surface_vtp_td_location()` method.
    """


class CustomVolumePointCloud(PostProcessing):
    """Provides a representation of a CustomVolumePointCloud post-processing.

    This class is generated through the :meth:`~PredictionPostProcessings.custom_volume_point_cloud()` method.
    """

    @property
    def data(self) -> DownloadableResult:
        """:class:`DownloadableResult` object that allows access to the custom volume VTP
        either directly in memory or by download it into a file.

        Accessing this property blocks until the data is ready
        """
        self.wait()
        results = self._get_results(cache=False)
        return DownloadableResult(results["data"]["resources"]["vtp"], self._client)


class PredictionPostProcessings:
    """Acts as the namespace inside :py:class:`~ansys.simai.core.data.predictions.Prediction` objects.

    This class allows you to analyze the results of a prediction.

    It can be accessed from any prediction object through its
    :attr:`~ansys.simai.core.data.predictions.Prediction.post` property:

    Example:
        .. code-block:: python

            sim = simai.predictions.get("<prediction_id>")
            # Run the global coefficients
            coefs = sim.post.global_coefficients()
    """

    def __init__(self, prediction: "Prediction"):
        self._client = prediction._client
        self.prediction = prediction
        # _post_processings contains, for each Type,
        # a Dict of params -> postProcess
        self._post_processings: Dict[Type[PostProcessing], Dict[frozenset, PostProcessing]] = {}

    def global_coefficients(self, run: bool = True) -> Optional[GlobalCoefficients]:
        """Compute or get the global coefficients of the prediction.

        This is a non-blocking method. It returns the ``GlobalCoefficients``
        object without waiting. This object may not have data right away
        if the computation is still in progress. Data is filled
        asynchronously once the computation is finished.
        The state of computation can be monitored with the ``is_ready`` flag
        or waited upon with the ``wait()`` method.

        Computation is launched only on first call of this method.
        Subsequent calls do not relaunch it.

        Args:
            run: Boolean indicating whether to compute or get the postprocessing.
                The default is ``True``. If ``False``, the postprocessing is not
                computed, and ``None`` is returned if it does not exist yet.

        Returns:
            ``GlobalCoefficients`` object that eventually contains
            the global coefficients with its pressure and velocity components.
            Returns ``None`` if ``run=False`` and the postprocessing does not exist.
        """
        return self._get_or_run(GlobalCoefficients, {}, run)

    def surface_evolution(
        self, axis: str, delta: float, run: bool = True
    ) -> Optional[SurfaceEvolution]:
        """Compute or get the SurfaceEvolution for specific parameters.

        This is a non-blocking method. It returns the ``SurfaceEvolution``
        object without waiting. This object may not have data right away
        if computation is still in progress. Data is filled
        asynchronously once the computation is finished.
        The state of computation can be monitored with the ``is_ready`` flag
        or waited upon with the ``wait()`` method.

        The computation is launched only on first call of this method
        with a specific set of parameters. Subsequent calls with the
        same parameters do not relaunch it.

        Args:
            axis: Axis to compute the surface evolution for.
            delta: Increment of the abscissa in meters.
            run: Boolean indicating whether to compute or get the postprocessing.
                The default is ``True``. If ``False``, the postprocessing is not
                computed, and ``None`` is returned if it does not exist yet.

        Returns:
            ``SurfaceEvolution`` that allows access to the values.
            Returns ``None`` if ``run=False`` and the postprocessing does not exist.
        """
        if axis not in ["x", "y", "z"]:
            raise TypeError("Axis must be x, y, or z.")
        if not isinstance(delta, numbers.Number) or not (delta > 0):
            raise TypeError(f"Delta must be a positive number (got: {delta}).")
        return self._get_or_run(SurfaceEvolution, {"axis": axis, "delta": delta}, run)

    def slice(
        self, axis: str, coordinate: float, format: str = "png", run: bool = True
    ) -> Optional[Slice]:
        """Compute or get a slice for specific plane parameters.

        This is a non-blocking method. It returns the ``Slice``
        object without waiting. This object may not have data right away
        if computation is still in progress. Data is filled
        asynchronously once the computation is finished.
        The state of computation can be monitored with the ``is_ready`` flag
        or waited upon with the ``wait()`` method.

        The computation is launched only on first call of this method
        with a specific set of parameters. Subsequent calls with the same
        parameters do not relaunch it.

        The slice is in the NPZ format.

        Args:
            axis: Axis to slice.
            coordinate: Coordinate along the given axis to slice at.
            format: Format of the output. The default is ``'png'``. Options
                are ``'png'`` and ``'vtp'``.
            run: Boolean indicating whether to compute or get the postprocessing.
                The default is ``True``. If ``False``, the postprocessing is not
                computed, and ``None`` is returned if it does not exist yet.

        Returns:
            ``Slice`` object that allows downloading the binary data.
            Returns ``None`` if ``run=False`` and the postprocessing does not exist.

        Example:
            Make a slice and open it in a new window using the `Pillow <https://pypi.org/project/pillow/>`_
            library.

            .. code-block:: python

                import ansys.simai.core
                from PIL import Image

                simai = ansys.simai.core.from_config()
                prediction = simai.predictions.list()[0]
                slice_data = prediction.post.slice("x", 50).data.in_memory()
                slice = Image.open(slice_data)
                slice.show()

        """
        if axis not in ["x", "y", "z"]:
            raise InvalidArguments(f"{axis} is not a valid axis. It should be x, y, or z.")
        if format not in ["png", "vtp"]:
            raise InvalidArguments(f"{format} is not a valid format. It should be png or vtp.")
        plane = convert_axis_and_coordinate_to_plane_eq_coeffs(axis, coordinate)
        return self._get_or_run(Slice, {"plane": plane, "output_format": format}, run)

    def surface_vtp(self, run: bool = True) -> Optional[SurfaceVTP]:
        """Compute or get the result of the prediction's surface in VTP format.

        This method associates all data with cells; if a variable is originally
        associated with points in the sample, it would be now associated with cells.

        It is a non-blocking method. It returns the ``PostProcessingVTP``
        object without waiting. This object may not have data right away
        if the computation is still in progress. Data is filled
        asynchronously once the computation is finished.
        The state of computation can be monitored with the ``is_ready`` flag
        or waited upon with the ``wait()`` method.

        The computation is launched only on first call of this method.
        Subsequent calls do not relaunch it.

        Args:
            run: Boolean indicating whether to compute or get the postprocessing.
                The default is ``True``. If ``False``, the postprocessing is not
                computed, and ``None`` is returned if it does not exist yet.

        Returns:
            :class:`SurfaceVTP` object that allows downloading the binary data.
            Returns ``None`` if ``run=False`` and the postprocessing does not exist.

        Examples:
            Run and download a surface VTP with data associated with cells.

            .. code-block:: python

                import ansys.simai.core

                simai = ansys.simai.core.from_config()
                prediction = simai.predictions.list()[0]
                surface_vtp = prediction.post.surface_vtp().data.download("/tmp/simai.vtp")

            Run a surface VTP with data association on cells, and open a plot using PyVista.

            .. code-block:: python

                import ansys.simai.core
                import pyvista
                import tempfile

                simai = ansys.simai.core.from_config()
                prediction = simai.predictions.list()[0]
                surface_vtp_data = prediction.post.surface_vtp().data
                # I don't want to save the file locally but pyvista doesn't read file-objects
                # Using temporary file as a workaround but a real path can be used instead
                with tempfile.NamedTemporaryFile(suffix=".vtp") as temp_vtp_file:
                    surface_vtp_data.download(temp_vtp_file.name)
                    surface_vtp = pyvista.read(temp_vtp_file.name)
                    surface_vtp.plot()
        """
        return self._get_or_run(SurfaceVTP, {}, run)

    def surface_vtp_td_location(self, run: bool = True) -> Optional[SurfaceVTPTDLocation]:
        """Compute or get the result of the prediction's surface in VTP format .

        This method keeps the original data association as they are in the sample.

        It is a non-blocking method. It returns the ``PostProcessingVTP``
        object without waiting. This object may not have data right away
        if the computation is still in progress. Data is filled
        asynchronously once the computation is finished.
        The state of computation can be monitored with the ``is_ready`` flag
        or waited upon with the ``wait()`` method.

        The computation is launched only on first call of this method.
        Subsequent calls do not relaunch it.

        Args:
            run: Boolean indicating whether to compute or get the postprocessing.
                The default is ``True``. If ``False``, the postprocessing is not
                computed, and ``None`` is returned if it does not exist yet.

        Returns:
            :class:`SurfaceVTPTDLocation` object that allows downloading the binary data.
            Returns ``None`` if ``run=False`` and the postprocessing does not exist.

        Examples:
            Run and download a surface VTP with the original data association.

            .. code-block:: python

                import ansys.simai.core

                simai = ansys.simai.core.from_config()
                prediction = simai.predictions.list()[0]
                surface_vtp = prediction.post.surface_vtp_td_location().data.download("/tmp/simai.vtp")

            Run a surface VTP with the original data association, and open a plot using PyVista.

            .. code-block:: python

                import ansys.simai.core
                import pyvista
                import tempfile

                simai = ansys.simai.core.from_config()
                prediction = simai.predictions.list()[0]
                surface_vtp_data = prediction.post.surface_vtp_td_location().data
                # I don't want to save the file locally but pyvista doesn't read file-objects
                # Using temporary file as a workaround but a real path can be used instead
                with tempfile.NamedTemporaryFile(suffix=".vtp") as temp_vtp_file:
                    surface_vtp_data.download(temp_vtp_file.name)
                    surface_vtp = pyvista.read(temp_vtp_file.name)
                    surface_vtp.plot()
        """
        return self._get_or_run(SurfaceVTPTDLocation, {}, run)

    def volume_vtu(self, run: bool = True) -> Optional[VolumeVTU]:
        """Compute or get the result of the prediction's volume in VTU format.

        This is a non-blocking method. It returns the ``PostProcessingVTU``
        object without waiting. This object may not have data right away
        if the computation is still in progress. Data is filled
        asynchronously once the computation is finished.
        The state of computation can be monitored with the ``is_ready`` flag
        or waited upon with the ``wait()`` method.

        The computation is launched only on first call of this method.
        Subsequent calls do not relaunch it.

        Args:
            run: Boolean indicating whether to compute or get the postprocessing.
                The default is ``True``. If ``False``, the postprocessing is not
                computed, and ``None`` is returned if it does not exist yet.

        Returns:
            :class:`VolumeVTU` object that allows downloading the binary data.

        Examples:
            Run and download a volume VTU

            .. code-block:: python

                import ansys.simai.core

                simai = ansys.simai.core.from_config()
                prediction = simai.predictions.list()[0]
                volume_vtu = prediction.post.volume_vtu().data.download("/tmp/simai.vtu")


            Run a volume VTU and open a plot using PyVista.

            .. code-block:: python

                import ansys.simai.core
                import pyvista
                import tempfile

                simai = ansys.simai.core.from_config()
                prediction = simai.predictions.list()[0]
                volume_vtu_data = prediction.post.volume_vtu().data
                # I don't want to save the file locally but pyvista doesn't read file-objects
                # Using temporary file as a workaround but a real path can be used instead
                with tempfile.NamedTemporaryFile(suffix=".vtu") as temp_vtu_file:
                    volume_vtu_data.download(temp_vtu_file.name)
                    volume_vtu = pyvista.read(temp_vtu_file.name)
                    volume_vtu.plot()
        """
        return self._get_or_run(VolumeVTU, {}, run)

    def custom_volume_point_cloud(self, run: bool = True) -> Optional[CustomVolumePointCloud]:
        """Compute or get the result of a predicted volume depending on the geometry's point cloud.

        This is a non-blocking method. It returns the :class:`CustomVolumePointCloud`
        object without waiting. This object may not have data right away
        if the computation is still in progress. Data is filled
        asynchronously once the computation is finished.
        The state of computation can be monitored with the ``is_ready`` flag
        or waited upon with the ``wait()`` method.

        The computation is launched only on first call of this method.
        Subsequent calls do not relaunch it.

        Args:
            run: Boolean indicating whether to compute or get the postprocessing.
                The default is ``True``. If ``False``, the postprocessing is not
                computed, and ``None`` is returned if it does not exist yet.

        Returns:
            :class:`CustomVolumePointCloud` object that allows downloading the binary data.

        Examples:
            Run and download a custom volume point cloud

            .. code-block:: python

                import ansys.simai.core

                simai = ansys.simai.core.from_config()
                prediction = simai.predictions.list()[0]
                prediction.geometry.upload_point_cloud("/data/my-point-cloud.vtp")
                custom_volume = prediction.post.custom_volume_point_cloud().data.download(
                    "/tmp/simai.vtp"
                )
        """
        if self.prediction.geometry.point_cloud is None:
            raise InvalidOperationError(
                """No point cloud file is attached to the geometry.
Attach the required file to enable CustomVolumePointCloud postprocessing."""
            )
        return self._get_or_run(CustomVolumePointCloud, {}, run)

    @property
    def _local_post_processings(self):
        """Postprocessings launched by the local session, which are waited upon."""
        local_post_processings = []
        for pp_by_type in self._post_processings.values():
            local_post_processings.extend(pp_by_type.values())
        return local_post_processings

    def _delete_local_post_processing(self, post_processing: PostProcessing):
        params = post_processing.fields.get("location", {})
        params_frozen = frozenset(params.items())
        del self._post_processings[post_processing.__class__][params_frozen]

    def _get_or_run(
        self, pp_class: Type[PostProcessing], params: Dict[str, Any], run: bool
    ) -> Optional[PostProcessing]:
        """Get the existing postprocessing or run one if it doesn't exist yet.

        Args:
            pp_class: Type of postprocessing.
            params: Parameters of the postprocessing.
            run: Boolean indicating whether to compute or get the postprocessing.
                If ``False``, this method only gets an existing postprocessing.

        This is a non-blocking method. It runs (if not already run orrunning) the postprocessing
        of given type with the given parameters. If ``run=False``, if a postprocessing already
        exits, it gets it.
        """
        # FIXME frozenset(params.items()) works as long as there are no
        # collision between params (axis and delta for surface evolution, param for slice)
        # but will be broken if a new type of postprocessings can have
        # two params with the same value.
        params_frozen = frozenset(params.items())
        # If a postprocessing of this type and with those params already exists locally, return it
        if pp_class in self._post_processings and params_frozen in self._post_processings[pp_class]:
            return self._post_processings[pp_class][params_frozen]
        if run:
            api_response = self._client._api.run_post_processing(
                self.prediction.id, pp_class._api_name(), params
            )
        else:
            api_response = self._client._api.get_post_processings_for_prediction(
                self.prediction.id, pp_class._api_name(), params
            )
            # The API is supposed to return a list of one element (because we filtered for the specific params).
            # Check that it's the case and extract that element from the list.
            if len(api_response) == 0:
                return None
            elif len(api_response) > 1:
                raise InvalidServerStateError(
                    cleandoc(
                        f"""
                        Multiple postprocessings were found when only one should be found.
                        {[pp["id"] for pp in api_response]}
                        """
                    )
                )
            else:
                api_response = api_response[0]

        post_processing = self._client._post_processing_directory._model_from(
            data=api_response, pp_class=pp_class, prediction=self.prediction
        )
        if pp_class not in self._post_processings:
            self._post_processings[pp_class] = {}
        self._post_processings[pp_class][params_frozen] = post_processing
        for location, warning_message in post_processing.fields.get("warnings", {}).items():
            logger.warning(f"{location}: {warning_message}")
        return post_processing

    def list(
        self,
        post_processing_type: Optional[Type[PostProcessing]] = None,
    ) -> List[PostProcessing]:
        """List the postprocessings associated with the prediction.

        See :func:`post_processings.list<ansys.simai.core.data.post_processings.PostProcessingDirectory.list>`
        """
        return self._client.post_processings.list(
            post_processing_type=post_processing_type,
            prediction=self.prediction,
        )


class PostProcessingDirectory(Directory[PostProcessing]):
    _data_model = PostProcessing

    @classmethod
    def _data_model_for_type_name(cls, pp_type_name: str) -> Optional[Type[PostProcessing]]:
        return getattr(sys.modules[__name__], pp_type_name, None)

    def _model_from(
        self,
        data: dict,
        pp_class: Optional[Type[PostProcessing]] = None,
        prediction: "Prediction" = None,
    ) -> PostProcessing:
        if pp_class is not None:
            constructor = pp_class
        else:
            constructor = self._data_model_for_type_name(data["type"])
        if not constructor:
            raise ValueError(f"""Received unknown postprocessing type {data["type"]}.""")
        post_processing = super()._model_from(data, data_model=constructor)
        post_processing._prediction = prediction
        return post_processing

    @property
    def info(self) -> List[Dict[str, Any]]:
        """Dictionary containing information about the available postprocessings and their parameters.

        Example:
            .. code-block:: python

                from pprint import pprint
                import ansys.simai.core

                simai = ansys.simai.core.from_config()
                post_processing_info = simai.post_processings.info
                pprint(post_processing_info)
        """
        return self._client.current_workspace.model_manifest.post_processings

    def get(self, id: str) -> PostProcessing:
        """Get a specific postprocessing object from the server.

        Args:
            id: ID of the postprocessing.

        Returns:
            :py:class:`PostProcessing` with the given ID if it exists.

        Raises:
            NotFoundError: No postprocessing with the given ID exists.
        """
        data = self._client._api.get_post_processing_result(id)
        return self._model_from(data)

    def list(
        self,
        post_processing_type: Optional[Type[PostProcessing]] = None,
        prediction: Optional[Identifiable["Prediction"]] = None,
        workspace: Optional[Identifiable["Workspace"]] = None,
    ) -> List[PostProcessing]:
        """List the postprocessings in the current workspace or associated with a prediction.

        Optionally you can choose to list only postprocessings of a specific type.
        For the name of the available postprocessings, see :ref:`available_pp`.

        Args:
            post_processing_type: Type of postprocessing to list.
            prediction: ID or :class:`model <.predictions.Prediction>` of a prediction.
                If a value is specified, only postprocessings associated with this prediction
                are returned.
            workspace: ID or :class:`model <.workspaces.Workspace>` of a workspace.
                If a value is specified, only postprocessings associated with this workspace
                are returned.

        Raises:
            NotFoundError: Postprocessing type, prediction ID or workspace ID are incorrect.
            InvalidArguments: If both prediction and workspace are specified.
            InvalidClientStateError: Neither prediction nor workspace are defined, default
                workspace is not set.

        Example:
            .. code-block:: python

                import ansys.simai.core

                simai = ansys.simai.core.from_config()
                prediction = simai.predictions.list()[0]
                post_processings = simai.post_processings.list(
                    ansys.simai.core.SurfaceEvolution, prediction.id
                )
        """
        pp_type_str = post_processing_type._api_name() if post_processing_type else None
        if workspace and prediction:
            raise InvalidArguments("Only one of Workspace or Prediction can be specified")
        if prediction:
            prediction_id = get_id_from_identifiable(prediction, required=False)
            post_processings = self._client._api.get_post_processings_for_prediction(
                prediction_id, pp_type_str
            )
        else:
            if workspace:
                workspace_id = get_id_from_identifiable(workspace, required=False)
            else:
                workspace_id = self._client.current_workspace.id
            post_processings = self._client._api.get_post_processings_in_workspace(
                workspace_id, pp_type_str
            )
        return list(map(self._model_from, post_processings))

    def run(
        self,
        post_processing_type: Union[str, Type[PostProcessing]],
        prediction: Identifiable["Prediction"],
        parameters: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> PostProcessing:
        """Run a postprocessing on a prediction.

        For the name and the parameters expected by the postprocessings,
        see :ref:`available_pp` and :ref:`pp_methods`. Note that the case
        of the class names must be respected.

        Args:
            post_processing_type: Type of postprocessing to run as a string
                or as the class itself.
            prediction: ID or :class:`model <.predictions.Prediction>` of the prediction
                to run the postprocessing for.
            parameters: Parameters to apply to the postprocessing, if needed.
                Alternatively, parameters can be passed as kwargs.
            **kwargs: Unpacked parameters for the postprocessing.

        Examples:
            .. code-block:: python

                import ansys.simai.core

                simai = ansys.simai.core.from_config()
                prediction = simai.predictions.list()[0]
                simai.post_processings.run(
                    ansys.simai.core.Slice, prediction, {"axis": "x", coordinate: 50}
                )

            Using kwargs:

            .. code-block:: python

                simai.post_processings.run(ansys.simai.core.Slice, prediction, axis="x", coordinate=50)
        """
        if isinstance(post_processing_type, str):
            post_processing_type = getattr(
                sys.modules[__name__], post_processing_type, post_processing_type
            )

        if isinstance(post_processing_type, type) and issubclass(
            post_processing_type, PostProcessing
        ):
            pp_class = post_processing_type
        else:
            raise InvalidArguments(
                cleandoc(
                    f""""{post_processing_type}" is not a valid postprocessing type.
                    You can find the available postprocessings by accessing the
                    ``.post_processings.info`` attribute of your SimAI client.
                    """
                )
            )
        prediction = self._client.predictions.get(get_id_from_identifiable(prediction))
        if not parameters:
            parameters = {}
        parameters.update(**kwargs)
        return prediction.post._get_or_run(pp_class, parameters, True)

    def delete(self, post_processing: Identifiable[PostProcessing]):
        """Delete a postprocessing.

        Args:
            post_processing: ID or :class:`model <PostProcessing>` of the postprocessing.
        """
        # FIXME?: This won't update the post_processings of the prediction's PredictionPostProcessings if any.
        # Doing so would require an extra call to get the prediction info and I'm not sure there's really a point
        self._client._api.delete_post_processing(get_id_from_identifiable(post_processing))



# Copyright (C) 2023 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import json
import logging
import os
import threading
from typing import Optional

from ansys.simai.core.api.mixin import ApiClientMixin
from ansys.simai.core.errors import ConnectionError
from ansys.simai.core.utils.configuration import ClientConfig
from ansys.simai.core.utils.sse_client import ReconnectingSSERequestsClient

logger = logging.getLogger(__name__)


SSE_ENDPOINT = "sessions/events"


class SSEMixin(ApiClientMixin):
    """Provides the client for the server-sent-events ("/sessions/events")."""

    def __init__(self, config: ClientConfig, simai_client=None):
        super().__init__(config=config)

        if simai_client:
            self.simai_client = simai_client
        else:
            logger.warning("SSEMixin has no SIM AI client.")

        # Disable sse thread in unit tests
        if config.no_sse_connection:
            logger.debug("Disabling SSE connection.")
            return

        # Flag for stopping the threads when this object is destroyed.
        # A flag is used because Python's threading library does not provide
        # a "stop" command.
        # The _stop_sse_threads is to allow unit tests to kill the thread
        # immediately without needing another event.
        self._stop_sse_threads = getattr(config, "_stop_sse_threads", False)
        logger.debug("Connecting to SSE.")

        def sse_connection_factory(last_event_id: Optional[str]):
            headers = {"Accept": "text/event-stream"}
            if last_event_id is not None:
                headers["Last-Event-ID"] = last_event_id
            return self._get(
                self._get_sse_url(),
                stream=True,
                headers=headers,
                return_json=False,
            )

        try:
            self.sse_client = ReconnectingSSERequestsClient(sse_connection_factory)
        except Exception as e:
            raise ConnectionError("Impossible to connect to event's endpoint.") from e
        logger.debug("SSEMixin is connected to SSE endpoint.")
        logger.debug("Starting listener thread.")
        self.listener_thread = threading.Thread(target=self._sse_thread_loop, daemon=True)
        self.listener_thread.start()
        logger.debug("Started listener thread.")

    def __del__(self):
        self._stop_sse_threads = True

    def _sse_thread_loop(self):
        try:
            for event in self.sse_client.events():
                if self._stop_sse_threads:
                    break
                self._handle_sse_event(event)
        except Exception:
            logger.critical("Unhandled exception in SSE thread: ", exc_info=True)
            # if object has been garbage collected, ignore exceptions
            if not self._stop_sse_threads:
                os._exit(1)

    def _handle_sse_event(self, event):
        try:
            if not event.data:
                # often a message has an empty data: ignore it
                return
            data = json.loads(event.data)
            logger.debug(f"received {data}")

            if "type" not in data:
                raise ValueError("No type is given for SSE event.")
            msg_type = data["type"]
            if msg_type == "session":
                self._handle_session_event(data)
            elif msg_type in ["job", "resource"]:
                self._handle_data_model_event(data)
        except Exception as e:
            msg = f"Error handling SSE event: {e}."
            logger.error(msg)

    def _handle_data_model_event(self, data):
        target = data["target"]
        if target["type"] == "geometry":
            self.simai_client.geometries._handle_sse_event(data)
        elif target["type"] == "prediction":
            self.simai_client.predictions._handle_sse_event(data)
        elif target["type"] == "post_processing":
            self.simai_client.post_processings._handle_sse_event(data)
        elif target["type"] == "training_data":
            self.simai_client.training_data._handle_sse_event(data)
        elif target["type"] == "training_data_part":
            self.simai_client.training_data_parts._handle_sse_event(data)
        elif target["type"] == "optimization":
            self.simai_client.optimizations._handle_sse_event(data)
        elif target["type"] == "optimization_trial_run":
            self.simai_client._optimization_trial_run_directory._handle_sse_event(data)
        elif target["type"] == "model":
            self.simai_client._model_directory._handle_sse_event(data)
        elif target["type"] == "project":
            if "action" in target:
                if target["action"] == "check":
                    self.simai_client._check_gc_formula_directory._handle_sse_event(data)
                elif target["action"] == "compute":
                    self.simai_client._compute_gc_formula_directory._handle_sse_event(data)
                else:
                    logger.debug(
                        f"Unknown action {target['action']} of type {target['type']} received for job or resource event. Ignoring."
                    )

        else:
            logger.debug(
                f"Unknown type {target['type']} received for job or resource event. Ignoring."
            )

    def _handle_session_event(self, data):
        # for now we don't handle session status
        service = data.get("service", "unknown")
        status = data.get("status", "unknown")
        logger.info(f"A session for {service} is now {status}")

    def _get_sse_url(self):
        return self.build_full_url_for_endpoint(SSE_ENDPOINT)



# Copyright (C) 2023 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
import os
import platform
from inspect import cleandoc
from pathlib import Path
from typing import Any, Dict, Optional

import tomli
from pydantic.v1.utils import deep_update

from ansys.simai.core.data.types import Path as PathType
from ansys.simai.core.errors import ConfigurationNotFoundError, InvalidConfigurationError

logger = logging.getLogger(__name__)


def _scan_defaults_config_paths() -> Optional[Path]:
    """Look for configuration files in the default locations.

    Returns:
        Path of the first configuration file from the list of existing configuration files.
    """
    system = platform.system()
    if system == "Windows":
        base_folder = Path(os.environ.get("APPDATA", "~")).expanduser()
        config_file = base_folder / "Ansys" / "simai.conf"
        if config_file.is_file():
            return config_file
    elif system == "Linux" or system == "Darwin":
        xdg_config = Path(os.getenv("XDG_CONFIG_HOME") or "~/.config")
        conf_paths = [
            xdg_config / "ansys_simai.conf",
            xdg_config / "ansys/simai.conf",
            "~/.ansys_simai.conf",
            "~/.ansys/simai.conf",
            "/etc/ansys_simai.conf",
            "/etc/ansys/simai.conf",
        ]
        for path in conf_paths:
            path = Path(path).expanduser()  # noqa: PLW2901
            if path.is_file():
                logger.debug(f"Found a configuration file at {path}.")
                return path
    else:
        raise ConfigurationNotFoundError("Could not determine OS.")
    # If no file was found
    return None


def get_config(
    path: Optional[PathType] = None,
    profile: str = "default",
    ignore_missing=False,
    **kwargs,
) -> Dict[Any, Any]:
    """Get configuration file, validate it, and return it as a flat dictionary.

    Args:
        path: Path of the configuration file. The default is ``None, in which
            case the method looks in default locations.
        profile: Profile to load. If no oath is specified, the method looks for ``[default]``.
        ignore_missing: Whether to raise an exception if no path to a configuration
            file is found. The default is ``False``.
        **kwargs: Overrides to apply to the configuration.
    """
    config_path = path or _scan_defaults_config_paths()
    if config_path is None:
        if ignore_missing:
            return kwargs
        raise ConfigurationNotFoundError
    config_path = Path(config_path).expanduser()
    if not os.path.isfile(config_path):
        raise ConfigurationNotFoundError
    with open(config_path, "rb") as f:
        try:
            all_config = tomli.load(f)
        except tomli.TOMLDecodeError as e:
            raise InvalidConfigurationError(f"Invalid configuration: {e}") from None
    config = all_config.get(profile, None)
    if config is None:
        raise InvalidConfigurationError(
            cleandoc(  # Takes care of the indentation
                f"""
                Did not find the [{profile}] profile section in the configuration file.
                Make sure that you have a '[default]' section or specify the name of the profile in your 'from_config' call.
                """
            )
        )
    config["_config_file_profile"] = profile
    # Apply overrides from kwargs if any
    config = deep_update(config, kwargs)
    return config



# Copyright (C) 2023 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
from numbers import Number
from typing import TYPE_CHECKING, Any, BinaryIO, Dict, List, Optional, Union

from ansys.simai.core.data.base import (
    ComputableDataModel,
    Directory,
    UploadableResourceMixin,
)
from ansys.simai.core.data.geometry_utils import (
    _geometry_matches_range_constraints,
    _sweep,
)
from ansys.simai.core.data.post_processings import CustomVolumePointCloud
from ansys.simai.core.data.types import (
    BoundaryConditions,
    File,
    Identifiable,
    MonitorCallback,
    NamedFile,
    Range,
    build_boundary_conditions,
    get_id_from_identifiable,
    get_object_from_identifiable,
    unpack_named_file,
)
from ansys.simai.core.data.workspaces import Workspace
from ansys.simai.core.errors import InvalidArguments, InvalidOperationError

if TYPE_CHECKING:
    from ansys.simai.core.data.predictions import Prediction

logger = logging.getLogger(__name__)


class Geometry(UploadableResourceMixin, ComputableDataModel):
    """Provides the local representation of a geometry object."""

    def __repr__(self) -> str:
        return f"<Geometry: {self.id}, {self.name}>"

    @property
    def name(self) -> str:
        """Name of the geometry."""
        return self.fields["name"]

    @property
    def metadata(self) -> Dict[str, Any]:
        """User-given key-value associated with the geometry."""
        return self.fields["metadata"]

    @property
    def creation_time(self) -> str:
        """Time when the geometry was created in a UTC ISO8601 format string."""
        return self.fields["creation_time"]

    @property
    def point_cloud(self) -> Optional[Dict[str, Any]]:
        """The attached point cloud file information if any."""
        return self.fields.get("point_cloud")

    def rename(self, name: str) -> None:
        """Change the name of the geometry.

        Args:
            name: New name to give to the geometry.

        Note:
            Only the stem part is modified. The file extension is immutable.
            If a file extension is provided, it must be the same as the original one.
            If the new filename already contains dots other than for the extension,
            the extension must be provided.
        """
        self._client._api.update_geometry(self.id, name=name)
        self.reload()

    def update_metadata(self, metadata: Dict[str, Union[str, Number, bool]]):
        """Change the metadata of the geometry.

        - New keys-values are added.
        - Existing keys-values are overwritten.
        - Other key-values are not changed.

        To delete a metadata, set it to ``None`` explicitly.

        Args:
            metadata: Dictionary with the new data.

        Examples:
            Add or update a metadata.

            .. code-block:: python

                geometry.update_metadata({"new_metadata": "value", "existing_metadata": 10})

            Remove all metadata.

            .. code-block:: python

               geometry.update_metadata({key: None for key in geometry.metadata.keys()})
        """
        self._client._api.update_geometry(self.id, metadata=metadata)
        self.reload()

    def delete(self) -> None:
        """Delete the geometry and its data from the server.

        All the objects associated with this geometry (predictions and postprocessings)
        are also deleted.

        See Also:
            :func:`GeometryDirectory.delete`
        """
        self._client._api.delete_geometry(self.id)

    def run_prediction(
        self, boundary_conditions: Optional[BoundaryConditions] = None, **kwargs
    ) -> "Prediction":
        """Run a new prediction or return an existing prediction.

        This is a non-blocking method. The prediction object is returned.
        This object may be incomplete if its computation is not finished,
        in which case the information is filled once the computation is complete.
        The state of the computation can be monitored with the prediction's ``is_ready``
        attribute or waited upon with its ``wait()`` method.

        To learn more about the expected boundary conditions in your workspace, you can use the
        ``simai.current_workspace.model.boundary_conditions`` or ``simai.predictions.boundary_conditions``,
        where ``ex`` is your `~ansys.simai.core.client.SimAIClient` object.

        Args:
            boundary_conditions: Boundary conditions to apply as a dictionary.
            **kwargs: Boundary conditions to pass as keyword arguments.

        Returns:
            Created prediction object.

        Raises:
            ProcessingError: If the server failed to process the request.

        Examples:
            .. code-block:: python

                simai = ansys.simai.core.from_config()
                geometry = simai.geometries.list()[0]
                geometry.run_prediction(dict(Vx=10.5, Vy=2))

            Use kwargs:

            .. code-block:: python

                prediction = geometry.run_prediction(Vx=10.5, Vy=2)
        """
        bc = build_boundary_conditions(boundary_conditions, **kwargs)
        prediction_response = self._client._api.run_prediction(self.id, boundary_conditions=bc)
        return self._client.predictions._model_from(prediction_response)

    def get_predictions(self) -> List["Prediction"]:
        """Get the prediction objects associated with the geometry."""
        predictions_data = self._client._api.get_geometry_predictions(self.id)
        return [self._client.predictions._model_from(pred_data) for pred_data in predictions_data]

    def download(
        self,
        file: Optional[File] = None,
        monitor_callback: Optional[MonitorCallback] = None,
    ) -> Union[None, BinaryIO]:
        """Download the geometry into the provided file or in memory if no file is provided.

        Args:
            file: Optional binary file-object or the path of the file to put the
                content into.
            monitor_callback: Optional callback to monitor the progress of the download.
                For more information, see the :obj:`~ansys.simai.core.data.types.MonitorCallback`
                object.

        Returns:
            ``None`` if a file is provided or the :class:`~io.BytesIO` object with the geometry's content otherwise.
        """
        return self._client._api.download_geometry(self.id, file, monitor_callback)

    def sweep(
        self,
        swept_metadata: Optional[Union[str, List[str]]] = None,
        fixed_metadata: Optional[List[str]] = None,
        geometries: Optional[List["Geometry"]] = None,
        order: Optional[int] = None,
        include_center: Optional[bool] = None,
        include_diagonals: Optional[bool] = None,
        tolerance: Optional[float] = None,
    ) -> List["Geometry"]:
        """Return geometries whose metadata are closest to the candidate geometry.

        This method returns geometries that have the values closest to the candidate
        geometry for each considered metadata variable. For example, if
        sweeping along ``length`` and ``width`` metadata variables, the method
        returns geometries that have identical width and the closest smaller
        and larger length, as well as identical length and the closest smaller
        and larger width.

        The ``fixed_metadata`` array allows you to fix one or several variables.
        For each fixed variable, the resulting geometries must have
        a metadata value equal to the considered geometry. For example, if
        ``fixed_metadata`` is ``["xbow"]``, every ``geometry.metadata["xbow"]``
        result must be equal to the ``candidate_geometry.metadata["xbow"]``.

        Metadata passed neither in ``swept_metadata`` nor in ``fixed_metadata``
        are ignored and can have any value (or absence thereof).

        Args:
            swept_metadata: Optional metadata name or a list of metadata names
                to consider. Only variables containing numerical values are
                supported. If no metadata names are passed, all metadata containing
                numerical values are taken into account.
            fixed_metadata: Optional list of metadata variables that should
                be fixed, meaning that all the resulting geometries
                have those values equal to the candidate geometry.
            geometries: Optional list of ``Geometry`` objects to consider for sweeping.
                If no ``Geometry`` objects are passed, all geometries are used.
            tolerance: Optional delta. If the difference between two numbers
                is lower than the tolerance, they are considered as equal.
                The default is ``10**-6``.
            order: Optional depth of the sweep. The default is ``1``. This parameter
                determines the number of returned groups of equal smaller and
                larger values for each swept variable. For example, if sweeping
                on a space with lengths ``[1, 2.1, 2.1, 3.5, 3.5, 4, 4]``
                from the candidate with ``length=1``, ``order=2`` returns
                the geometries with lengths ``2.1, 2.1, 3.5, 3.5``.
            include_center: Optional Boolean indicating whether geometries with values
                equal to the candidate geometry (including the candidate itself) are
                to be returned among the result. The default is ``False``.
            include_diagonals: Optional Boolean indicating whether to include diagonals
                when sweeping on more than one variable. The default is ``False``.
                For example, if sweeping on two variables from point ``(0, 0)``
                and with ``order=1``, in addition to ``(0, 1)`` and ``(1, 0)``,
                geometry ``(1, 1)`` is returned.

        Returns:
            List of ``Geometry`` objects neighboring the candidate geometry for each metadata.

        Raises:
            ValueError: If a passed variable doesn't exist in the
                candidate geometry.
            ValueError: If the considered metadata contains non-numerical values
                or mixed numerical and non numerical values.

        Example:
            .. code-block:: python

                import ansys.simai.core

                simai = ansys.simai.core.from_config()
                geom = simai.geometries.get("kz19jyqm")
                geometries = geom.sweep(["length"])
        """
        if geometries is None:
            geometries = self._client.geometries.list()
        return _sweep(
            candidate_geometry=self,
            geometries=geometries,
            swept_metadata=swept_metadata,
            fixed_metadata=fixed_metadata,
            order=order,
            include_center=include_center,
            include_diagonals=include_diagonals,
            tolerance=tolerance,
        )

    def upload_point_cloud(
        self, file: NamedFile, monitor_callback: Optional[MonitorCallback] = None
    ):
        """Upload a point cloud for this geometry.

        Only the vtp file format is supported

        Args:
            file: :obj:`~ansys.simai.core.data.types.NamedFile` to upload.
            monitor_callback: Optional callback for monitoring the progress of the upload.
                For more information, see the :obj:`~ansys.simai.core.data.types.MonitorCallback`
                object.
        """
        with unpack_named_file(file) as (readable_file, name, extension):
            response = self._client._api.create_point_cloud(self.id, name, extension)
            upload_id = response["upload_id"]
            point_cloud_fields = response["point_cloud"]
            parts = self._client._api.upload_parts(
                f"point-clouds/{point_cloud_fields['id']}/part",
                readable_file,
                upload_id,
                monitor_callback=monitor_callback,
            )
            self._client._api.complete_point_cloud_upload(
                point_cloud_fields["id"], upload_id, parts
            )
            self._fields["point_cloud"] = point_cloud_fields

    def delete_point_cloud(self):
        """Delete the associated point cloud file."""
        if not self.point_cloud:
            raise InvalidOperationError("No point cloud file to delete")
        self._client._api.delete_point_cloud(self.point_cloud["id"])
        self._fields["point_cloud"] = None
        # The post-processing will be deleted server side but the prediction cache will still
        # have the post-processing registered. Meaning a new _get_or_run of the post-processing will
        # attempt a get, thinking the post-processing exists. So for all the registered prediction who
        # belong to the current geometry, we need to clear the CustomVolumePointCloud local registry.
        for prediction in self._client._prediction_directory._registry.values():
            if prediction.geometry_id == self.id:
                prediction.post._post_processings[CustomVolumePointCloud] = {}


class GeometryDirectory(Directory[Geometry]):
    """Provides a collection of methods related to geometries.

    This class is accessed through ``client.geometries``.

    Example:
        .. code-block:: python

            import ansys.simai.core

            simai = ansys.simai.core.from_config()
            simai.geometries.list()
    """

    _data_model = Geometry

    def list(
        self,
        workspace: Optional[Identifiable[Workspace]] = None,
        filters: Optional[Dict[str, Union[str, float, Range]]] = None,
    ) -> List[Geometry]:
        """List geometries from the server that belong to the currently set workspace or the specified one.

        Args:
            workspace: ID or :class:`model <.workspaces.Workspace>` of the workspace to list geometries for.
                This parameter is required if no global workspace is set for the client.
            filters: Optional filters. Only the elements with matching key-values in
                their metadata are returned. Each filter can be one of the following data types:

                - A string
                - A numerical value (int or float)
                - A :py:class:`Range` condition for filtering values matching a
                  given numerical range of values

        Returns:
            List of all or filtered geometries on the server.

        Raises:
            TypeError: If a :py:class:`Range` condition is applied on non-numerical metadata.
        """
        workspace_id = get_id_from_identifiable(workspace, default=self._client._current_workspace)

        # for now we filter ranges locally, so we filter
        # local filters (ranges) from server filters (everything else)
        ranges: Dict[str, Range] = {}
        server_filters: Dict[str, Union[str, float]] = {}
        if filters is not None:
            for metadata, filter in filters.items():
                if isinstance(filter, Range):
                    ranges[metadata] = filter
                else:
                    server_filters[metadata] = filter
        geometries = [
            self._model_from(data)
            for data in self._client._api.geometries(workspace_id, server_filters)
        ]
        return [
            geometry
            for geometry in geometries
            if _geometry_matches_range_constraints(geometry, ranges)
        ]

    def filter(self, **kwargs: Dict[str, Union[str, float, Range]]) -> List[Geometry]:
        """Filter geometries from the server that belong to the currently set workspace.

        Args:
            kwargs: Filters to apply. Only the elements with matching key-values in
                their metadata are returned. Each filter can be one of the following data types:

                - A string
                - A numerical value (int or float)
                - A :py:class:`Range` condition for filtering values matching a
                  given numerical range of values

        Returns:
            List of filtered geometries on the server.

        Raises:
            TypeError: If a :py:class:`Range` condition is applied on non-numerical metadata.
        """
        return self.list(filters=kwargs)

    def get(
        self,
        name: Optional[str] = None,
        id: Optional[str] = None,
        workspace: Optional[Identifiable[Workspace]] = None,
    ) -> Geometry:
        """Get a specific geometry object from the server either by name or ID.

        You can specify either the ID or the name, not both.


        Args:
            name: Name of the geometry.
            id: ID of the geometry.
            workspace: ID or :class:`model <.workspaces.Workspace>` of the workspace containing the geometry.
                This parameter is necessary if providing a name and no global workspace is set for the client.

        Returns:
            :py:class:`Geometry`.

        Raises:
            InvalidArguments: If neither a name nor an ID is given.
            NotFoundError: If no geometry with the given name or ID exists.

        Examples:
            Get a geometry by name.

            .. code-block:: python

                import ansys.simai.core

                simai = ansys.simai.core.from_config()
                geometry = simai.geometries.get("my_geometry.stl")
                # geometry = simai.geometries.get(name="my_geometry.stl") # is equivalent

            Get a geometry by ID.

            .. code-block:: python

                    geometry = simai.geometries.get(id="abcdef12")
        """
        if name and id:
            raise InvalidArguments("Name and ID cannot both be specified.")
        if name:
            return self._model_from(
                self._client._api.get_geometry_by_name(
                    name,
                    get_id_from_identifiable(workspace, default=self._client._current_workspace),
                )
            )
        if id:
            return self._model_from(self._client._api.get_geometry(id))
        raise InvalidArguments("Either the name or the ID must be specified.")

    def delete(self, geometry: Identifiable[Geometry]) -> None:
        """Delete a specific geometry and its data from the server.

        All the objects associated with this geometry (predictions and postprocessings)
        are also deleted.

        Args:
            geometry: ID or :class:`model <Geometry>` of the geometry.

        Raises:
            NotFoundError: No geometry with the given ID exists.

        See Also:
            :func:`Geometry.delete`
        """
        self._client._api.delete_geometry(get_id_from_identifiable(geometry))

    def upload(  # noqa: D417
        self,
        file: NamedFile,
        metadata: Optional[Dict[str, Any]] = None,
        workspace: Optional[Identifiable[Workspace]] = None,
        monitor_callback: Optional[MonitorCallback] = None,
        **kwargs,
    ) -> Geometry:
        """Upload a geometry to the SimAI platform.

        Args:
            file: Binary file-object or the path of the geometry to open.
                For more information, see the :class:`~ansys.simai.core.data.types.NamedFile` class.
            metadata: Optional metadata to add to the geometry's simple key-value store.
                Lists and nested objects are not supported.
            workspace: ID or :class:`model <.workspaces.Workspace>` of the workspace to
                upload the geometry to. This parameter is only necessary if no workspace
                is set for the client.
            monitor_callback: Optional callback for monitoring the progress of the download.
                For more information, see the :obj:`~ansys.simai.core.data.types.MonitorCallback`
                object.

        Returns:
            Created :py:class:`Geometry` object.
        """
        workspace = get_object_from_identifiable(
            workspace, self._client.workspaces, default=self._client._current_workspace
        )
        with unpack_named_file(file) as (readable_file, name, extension):
            if extension not in workspace.model_manifest.geometry["accepted_file_formats"]:
                raise InvalidArguments(
                    f"Got a file with {extension} extension but expected one of : {workspace.model_manifest.geometry['accepted_file_formats']}"
                )
            (geometry_fields, upload_id) = self._client._api.create_geometry(
                workspace.id, name, extension, metadata
            )
            geometry = self._model_from(geometry_fields, is_upload_complete=False)
            parts = self._client._api.upload_parts(
                f"geometries/{geometry.id}/part",
                readable_file,
                upload_id,
                monitor_callback=monitor_callback,
            )
            self._client._api.complete_geometry_upload(geometry.id, upload_id, parts)
        return geometry

    def download(
        self,
        geometry: Identifiable[Geometry],
        file: Optional[File] = None,
        monitor_callback: Optional[MonitorCallback] = None,
    ) -> Union[None, BinaryIO]:
        """Download the geometry with the given ID into the file at the given path.

        Args:
            geometry: ID or :class:`model <Geometry>` of the geometry.
            file: Optional binary file-object or the path of the file to put the
                content into.
            monitor_callback: Optional callback for monitoring the progress of the download.
                For more information, see the :obj:`~ansys.simai.core.data.types.MonitorCallback`
                object.

        Returns:
            ``None`` if a file is provided or a :class:`~io.BytesIO` object with the geometry's content otherwise.

        See Also:
            :func:`Geometry.download`
        """
        return self._client._api.download_geometry(
            get_id_from_identifiable(geometry), file, monitor_callback
        )

    def sweep(
        self,
        candidate_geometry: Identifiable[Geometry],
        swept_metadata: Optional[Union[str, List[str]]] = None,
        fixed_metadata: Optional[List[str]] = None,
        geometries: Optional[List["Geometry"]] = None,
        order: Optional[int] = None,
        include_center: Optional[bool] = None,
        include_diagonals: Optional[bool] = None,
        tolerance: Optional[float] = None,
    ) -> List["Geometry"]:
        """Get the geometries whose metadata are closest to the candidate geometry.

        For more information, see the :func:`Geometry.sweep` method.

        Example:
            .. code-block:: python

                import ansys.simai.core

                simai = ansys.simai.core.from_config()
                geom = simai.geometries.get("kz19jyqm")
                geometries = simai.geometries.sweep(geom, ["length"])

        See Also:
            :func:`Geometry.sweep`
        """
        candidate_geometry = get_object_from_identifiable(
            candidate_geometry, self._client.geometries
        )
        return candidate_geometry.sweep(
            swept_metadata=swept_metadata,
            fixed_metadata=fixed_metadata,
            geometries=geometries,
            order=order,
            include_center=include_center,
            include_diagonals=include_diagonals,
            tolerance=tolerance,
        )



# Copyright (C) 2023 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
from abc import ABC, abstractmethod
from threading import Event
from typing import TYPE_CHECKING, Dict, Generic, Optional, Type, TypeVar

from ansys.simai.core.errors import ProcessingError

if TYPE_CHECKING:
    import ansys.simai.core.client


logger = logging.getLogger(__name__)


ERROR_STATES = ["failure", "rejected", "cancelled"]
PENDING_STATES = [
    "unknown",
    "queued",
    "pending request",
    "requested",
    "processing",
    "pending_retry",
]


class DataModel:
    """A base class representing a single object of a particular type on the server."""

    def __init__(
        self,
        client: "ansys.simai.core.client.SimAIClient",
        directory: "Directory",
        fields: dict,
        **kwargs,
    ):
        # Override id_key to configure the field to return with :py:meth:`Model.id`.
        self.id_key = "id"
        self.fields = fields or {}
        self._directory = directory
        self._client = client

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self.id}>"

    @property
    def id(self) -> str:
        """ID of the object on the server."""
        # Uses the :py:attr:`id_key` to determine the name of the id.
        # :py:attr:`id_key` defaults to id
        return self._fields[self.id_key]

    @property
    def fields(self) -> dict:
        """Dictionary containing the raw object representation."""
        return self._fields

    @fields.setter
    def fields(self, new_fields: dict) -> None:
        self._fields = new_fields

    def reload(self) -> None:
        """Refresh the object with its representation from the server."""
        updated_data = self._directory.get(id=self.id)
        self.fields = updated_data._fields

    @property
    def _classname(self):
        return self.__class__.__name__


class ComputableDataModel(DataModel):
    """Provides the base class for all computable models whose creation eventually succeeds or fails."""

    def __init__(
        self,
        client: "ansys.simai.core.client.SimAIClient",
        directory: "Directory",
        fields: dict,
        **kwargs,
    ):
        super().__init__(client=client, directory=directory, fields=fields)

        # threading.Event object that represents the finalization state
        # of this object. Its wait() will block until the object
        # is finished processing. Used by this class' wait() method.
        # It is `cleared` state as long as the object is loading (queud or processing)
        # and in `set` state when finalized (complete or failed)
        self._is_over = Event()

        # set the initial state of _is_over event,
        # so the user can wait() on the object
        # even before we receive any SSE event
        state = self.fields.get("state")
        if state == "successful" or state in ERROR_STATES:
            logger.debug(f"set {self._classname} {fields['id']} as idle")
            self._set_is_over()
        else:
            logger.debug(f"set {self._classname} {fields['id']} as pending")
            self._set_is_pending()

        self._has_failed = state in ERROR_STATES

    def _set_is_pending(self):
        # todo maybe the is_failed status should be cleared here?
        # if is it possible to re-launch a failed operation?
        self._is_over.clear()

    @property
    def is_pending(self):
        """Boolean indicating if the object is still in creation.
        The value becomes ``False`` once object creation is either successful
        or has failed.

        See Also:
            - :meth:`~wait`
            - :attr:`~is_ready`
            - :attr:`~has_failed`
        """
        return not self._is_over.is_set()

    def _set_has_failed(self):
        self._has_failed = True
        self._set_is_over()

    @property
    def has_failed(self):
        """Boolean indicating if the creation of the object failed.

        See Also:
            - :attr:`~failure_reason`
            - :meth:`~wait`
            - :attr:`~is_ready`
            - :attr:`~is_pending`
        """
        return self._has_failed

    @property
    def is_ready(self):
        """Boolean indicating if the object has finished creating without error.

        See Also:
            - :meth:`~wait`
            - :attr:`~is_pending`
            - :attr:`~has_failed`
        """
        return not self.is_pending and not self.has_failed

    @property
    def failure_reason(self):
        """Optional message giving the causes for why the
        creation of the object failed.

        See Also:
            - :attr:`~has_failed`
        """
        return self.fields.get("error")

    @property
    def _failure_message(self):
        reason_str = f": {self.fields['error']}" if self.failure_reason else ""
        return f"{self._classname} id {self.id} failed {reason_str}"

    def _set_is_over(self):
        """Set the object as idle, that is without loading
        (either because loading finished or failed),
        meaning a wait() returns immediately.
        """
        self._is_over.set()

    def wait(self, timeout: Optional[float] = None) -> bool:
        """Wait for all jobs concerning the object to either finish
        or fail.

        Args:
            timeout: Maximum amount of time in seconds to wait. The default is
                ``None``, in means that there is no maximum on the time to wait.

        Returns:
            ``True`` if the computation has finished, ``False`` if the operation timed out.
        """
        is_done = self._is_over.wait(timeout)
        if self.has_failed:
            raise ProcessingError(self._failure_message)
        return is_done

    def _unregister(self):
        self._directory._unregister_item(self)

    def _handle_job_sse_event(self, data):
        """Update object with the information and state received through the SSE."""
        logger.debug(f"Handling SSE job event for {self._classname} id {self.id}")
        self.fields = data["record"]
        state: str = data["record"]["state"]
        if state in PENDING_STATES:
            logger.debug(f"{self._classname} id {self.id} set status pending")
            self._set_is_pending()
        elif state == "successful":
            logger.debug(f"{self._classname} id {self.id} set status successful")
            self._set_is_over()
        elif state in ERROR_STATES:
            reason_str = f"with reason {self.fields.get('error')}"
            logger.debug(f"{self._classname} id {self.id} set status failed {reason_str}")
            self._set_has_failed()
            logger.error(self.fields.get("error"))


DataModelType = TypeVar("DataModelType", bound=DataModel)


class Directory(ABC, Generic[DataModelType]):
    _data_model: Type[DataModelType]

    def __init__(self, client: "ansys.simai.core.client.SimAIClient"):
        self._client = client

        # Registry for known objects of this type that have been created
        # locally. It is used to ensure that only one instance of a particular class
        # exists with a given ID.
        # Note that this dictionary must retain all models created locally
        # without relying on the user to keep a reference on them
        # (meaning for example that WeakValueDictionary is not appropriate),
        # as the use expects simai.wait() to work on any object, even those
        # with no explicit reference.
        self._registry: Dict[str, DataModel] = {}

    @abstractmethod
    def get(self, id: str) -> DataModel:  # noqa: D102
        pass

    def _model_from(
        self, data: dict, data_model: Optional[DataModelType] = None, **kwargs
    ) -> DataModelType:
        # _model_from overrides object data (fields),
        # thus it is, and should, only be called with
        # complete data from the server, such as from a GET request.
        if "id" not in data:
            raise ValueError("Cannot instantiate data object without an id")
        item_id = data["id"]
        if item_id in self._registry:
            # An instance with this ID already exists: Update and return it,
            # so there are not two objects with different data.
            item = self._registry[item_id]
            # update with new data
            item.fields = data
        else:
            constructor = data_model if data_model else self._data_model
            # item was unknown: create it and save it to the registry
            item = constructor(self._client, self, data, **kwargs)
            self._registry[item_id] = item
        return item

    def _unregister_item(self, item: DataModelType):
        """Remove the item from the internal registry,
        mainly after a deletion.
        """
        self._unregister_item_with_id(item.id)

    def _unregister_item_with_id(self, item_id: str):
        """Remove the item from the internal registry,
        mainly after a deletion.
        """
        item = self._registry.pop(item_id, None)
        if isinstance(item, ComputableDataModel):
            item._set_is_over()

    def _handle_sse_event(self, data):
        item_id: str = data["target"]["id"]
        if item_id not in self._registry:
            # We ignore object whose id is unknown:
            # we only listen to objects that have been
            # created by a local operation
            logger.debug(f"Ignoring event for unknown object id {item_id}")
            return
        item = self._registry[item_id]

        if data["type"] == "job" and isinstance(item, ComputableDataModel):
            item._handle_job_sse_event(data)
        elif data["type"] == "resource" and isinstance(item, UploadableResourceMixin):
            item._handle_resource_sse_event(data)
        else:
            logger.debug("Received a server update that could not be interpreted.")

    def _all_objects(self):
        return self._registry.values()


class UploadableResourceMixin:
    """Provides the class used to expand ``DataModel`` with support for a resource type message from SSE."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._is_upload_complete = Event()
        if kwargs.get("is_upload_complete", True):
            self._is_upload_complete.set()

    def _wait_for_upload_completion(self):
        self._is_upload_complete.wait()

    def _handle_resource_sse_event(self, data):
        state = data["status"]
        if state == "created":
            self._is_upload_complete.set()
        elif state == "upload_failed":
            self._is_upload_complete.set()
            ProcessingError(
                f"Could not complete upload because: {data.get('reason', 'Upload failed')}"
            )
        else:
            logger.error("Invalid resource state.")



# Copyright (C) 2023 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import io
import os
import pathlib
from contextlib import contextmanager
from enum import Enum
from numbers import Number
from typing import (
    Any,
    BinaryIO,
    Callable,
    Dict,
    Generator,
    List,
    Literal,
    Optional,
    Protocol,
    Tuple,
    TypedDict,
    TypeVar,
    Union,
)

from requests import Response

from ansys.simai.core.data.base import DataModel, DataModelType, Directory
from ansys.simai.core.errors import InvalidArguments
from ansys.simai.core.utils.files import file_path_to_obj_file
from ansys.simai.core.utils.numerical import (
    is_bigger_or_equal_with_tolerance,
    is_equal_with_tolerance,
    is_number,
    is_smaller_or_equal_with_tolerance,
    validate_tolerance_parameter,
)

BoundaryConditions = Dict[str, Number]
"""
:obj:`BoundaryConditions` describes the external conditions of a prediction.
"""

Path = Union[pathlib.Path, str, os.PathLike]
"""
Path to a file or folder as an :obj:`pathlib.Path` object or a format supported by ``pathlib``.
"""

File = Union[BinaryIO, io.RawIOBase, io.BufferedIOBase, Path]
"""
Either a binary file-object (:obj:`typing.BinaryIO`) or a :obj:`Path`.
"""

NamedFile = Union[Path, Tuple[File, str]]
"""
A named file is either a ``FilePath``, from which a name can be inferred, or a tuple with a file and a name.
To be valid, the name needs to contain an extension.

Example:
    .. code-block:: python

        file = "/path/to/my/file.stl"  # The file is named file.stl
        file = ("/path/to/my/file.stl", "override.stl")  # The file is named override.stl
        file = (io.BinaryIO(my_data), "file.stl")  # The file is named file.stl

        invalid_file = io.BinaryIO(my_data)  # Cannot infer name from binaryIO
        invalid_file = "/path/to/my/file"  # Inferred name has no extension
        invalid_file = ("/path/to/my/file.stl", "override")  # Specified name has no extension
"""


JSON = Union[Dict[str, Any], List[Dict[str, Any]]]
APIResponse = Union[Response, JSON]


MonitorCallback = Callable[[int], None]
"""Callback used to monitor the download or upload of a file.

For downloads, the callback is called one time with the total size of the download.
Subsequent calls are passed the number of bytes read in this iteration.

For uploads, the callback receives the number of bytes written each iteration.
"""

Identifiable = Union[DataModelType, str]
"""Either a model or the string ID of an object of the same type."""

D = TypeVar("D", bound=DataModel)
T_co = TypeVar("T_co", covariant=True)


def build_boundary_conditions(boundary_conditions: Optional[Dict[str, Number]] = None, **kwargs):
    bc = boundary_conditions if boundary_conditions else {}
    bc.update(**kwargs)
    if bc is None:
        raise ValueError("No boundary condition was specified.")
    if not is_boundary_conditions(bc):
        raise ValueError("Boundary conditions must be in a dictionary with numbers as values.")
    return bc


def is_boundary_conditions(bc):
    return isinstance(bc, dict) and all(is_number(x) for x in bc.values())


def are_boundary_conditions_equal(
    left: BoundaryConditions,
    right: BoundaryConditions,
    tolerance: Optional[Number] = None,
):
    if not is_boundary_conditions(left):
        raise TypeError(
            f"is_boundary_conditions_equal called with incorrect left parameter (received {left})"
        )
    if not is_boundary_conditions(right):
        raise TypeError(
            f"is_boundary_conditions_equal called with incorrect right parameter (received {right})"
        )
    validate_tolerance_parameter(tolerance)
    if left.keys() != right.keys():
        return False
    for key in left.keys():  # noqa: SIM118
        if not is_equal_with_tolerance(left[key], right[key], tolerance=tolerance):
            return False
    return True


class Range:
    """Describes a numerical range used for filtering geometries.

    Range objects describe a numerical range between a minimum and
    a maximum boundary. Both are optional. Thus, if no maximum boundary
    is passed, the range describes values greater than or equal to the
    minimum boundary. Note that ranges are inclusive. Thus, both minimum
    and maximum boundaries match if they are equal to the passed value
    (as opposed to Python's ``range()`` method).

    Ranges can be used as a filter in the
    :func:`geometries.list<ansys.simai.core.data.geometries.GeometryDirectory.list>` method.

    Args:
        min: Minimum boundary.
        max: Maximum boundary.
        tolerance: Tolerance delta. Two values whose difference is smaller
            than the tolerance are considered as equal.
    """

    def __init__(
        self,
        min: Optional[float] = None,
        max: Optional[float] = None,
        tolerance: Optional[float] = None,
    ):
        self.min = min
        self.max = max
        self.tolerance = tolerance

    def match_value(self, value: float) -> bool:
        """Determine whether the given value belongs to the :class:`Range` class."""
        if not is_number(value):
            return False
        # if min, value >= min
        if self.min is not None and not is_bigger_or_equal_with_tolerance(
            value, self.min, tolerance=self.tolerance
        ):
            return False
        # if max, value <= max
        return self.max is None or is_smaller_or_equal_with_tolerance(
            value, self.max, tolerance=self.tolerance
        )


class _HollowRange(Range):
    """_HollowRange is a range that excludes a value in its center."""

    def __init__(
        self,
        min: Optional[float] = None,
        max: Optional[float] = None,
        tolerance: Optional[float] = None,
        excluded_value: Optional[float] = None,
    ):
        super().__init__(min=min, max=max, tolerance=tolerance)
        self.excluded_value = excluded_value

    def match_value(self, value: float):
        if not super().match_value(value):
            return False
        return self.excluded_value is None or not is_equal_with_tolerance(
            value, self.excluded_value, tolerance=self.tolerance
        )


@contextmanager
def unpack_named_file(
    named_file: NamedFile,
) -> Generator[Tuple[BinaryIO, str, str], None, None]:
    """Unpack a named file by providing a readable file, its name, and an extension."""
    if (
        isinstance(named_file, Tuple)
        and len(named_file) == 2
        # == isinstance(named_file, File) on python >=3.10
        and isinstance(named_file[0], File.__args__)
    ):
        file = named_file[0]
        filename = named_file[1]
    # == isinstance(named_file, Path) on python >=3.10
    elif isinstance(named_file, Path.__args__):
        file = named_file
        filename = pathlib.Path(named_file).name
    else:
        raise InvalidArguments("Did not receive a valid named file type.")

    # Parse name and extension
    try:
        obj_name, file_ext = filename.rsplit(".", 1)
        assert file_ext
    except (ValueError, AssertionError):
        raise AttributeError(f"Could not determine file extension for {named_file}.") from None

    # Open the file if needed
    close_file = False
    if not isinstance(file, (io.RawIOBase, io.BufferedIOBase)):
        file = file_path_to_obj_file(file, "rb")
        close_file = True

    try:
        yield file, obj_name, file_ext
    finally:
        if close_file is True:
            file.close()


def get_id_from_identifiable(
    identifiable: Optional[Identifiable["D"]] = None,
    required: bool = True,
    default: Optional[Identifiable["D"]] = None,
) -> Optional[str]:
    if isinstance(identifiable, DataModel):
        return identifiable.id
    elif isinstance(identifiable, str):
        return identifiable
    elif default:
        return get_id_from_identifiable(default, required)
    elif required:
        raise InvalidArguments(f"Argument {identifiable} is neither a data model nor an ID string.")


def get_object_from_identifiable(
    identifiable: Optional[Identifiable],
    directory: Directory,
    default: Optional[Identifiable] = None,
) -> DataModel:
    if isinstance(identifiable, DataModel):
        return identifiable
    elif isinstance(identifiable, str):
        return directory.get(id=identifiable)
    elif default:
        return get_object_from_identifiable(default, directory)
    else:
        raise InvalidArguments(f"Argument {identifiable} is neither a data model nor an ID string.")


class SubsetEnum(str, Enum):
    """:obj:`SubsetEnum`: enumeration of possible subsets TrainingData can belong to if any."""

    TRAINING = "Training"
    TEST = "Test"


class RawSingleFilter(TypedDict):
    field: str
    operator: "FilterOperator"
    value: Any


FilterOperator = Literal["EQ", "LIKE", "IN", "GT", "GTE", "LT", "LTE"]
RawFilters = list[RawSingleFilter]
Filters = Union[dict[str, Any], list[tuple[str, FilterOperator, Any]], RawFilters]
"""
Filters type for ``list()`` endpoints that support them.

The simplified ``Filters`` syntax (only for ``EQ`` filters) is: ``{"name": "my-experiment"}``.
The full syntax is: ``[("name", "LIKE", "%thingy%"), ("size", "GT", 9000)]``.
In both cases, the conditions are ``AND`` together.
"""


def to_raw_filters(filters: Optional["Filters"]) -> Optional["RawFilters"]:
    if isinstance(filters, dict):
        return [{"field": k, "operator": "EQ", "value": v} for k, v in filters.items()]
    if isinstance(filters, list) and len(filters) > 0:
        if isinstance(filters[0], dict):
            return filters
        if isinstance(filters[0], (tuple, list)):
            return [{"field": fld, "operator": op, "value": val} for fld, op, val in filters]


class SizedIterator(Protocol[T_co]):
    """Alias of ``Iterator[T] & Sized``."""

    def __len__(self) -> int: ...
    def __next__(self) -> T_co: ...



# Copyright (C) 2023 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import json
from typing import TYPE_CHECKING, Any, Dict, Optional
from urllib.parse import urlencode

from ansys.simai.core.api.mixin import ApiClientMixin
from ansys.simai.core.utils.pagination import PaginatedAPIRawIterator

if TYPE_CHECKING:
    from ansys.simai.core.data.types import RawFilters


class TrainingDataClientMixin(ApiClientMixin):
    def delete_training_data(self, id: str) -> None:
        self._delete(f"training-data/{id}", return_json=False)

    def get_training_data(self, id: str) -> Dict[str, Any]:
        return self._get(f"training-data/{id}")

    def iter_training_data(self, filters: Optional["RawFilters"]) -> PaginatedAPIRawIterator:
        query = urlencode(
            [("filter[]", json.dumps(f, separators=(",", ":"))) for f in (filters or [])]
        )
        return PaginatedAPIRawIterator(self, f"training-data?{query}")

    def create_training_data(self, name: str, project_id: Optional[str] = None) -> Dict[str, Any]:
        args = {"name": name}
        if project_id:
            args["project"] = project_id
        return self._post(
            "training-data/",
            json=args,
        )

    def add_training_data_to_project(self, training_data_id: str, project_id: str):
        return self._put(
            f"training-data/{training_data_id}/project/{project_id}/association",
            return_json=False,
        )

    def remove_training_data_from_project(self, training_data_id: str, project_id: str):
        return self._delete(
            f"training-data/{training_data_id}/project/{project_id}/association",
            return_json=False,
        )

    def compute_training_data(self, training_data_id: str) -> None:
        self._post(f"training-data/{training_data_id}/compute")

    def get_training_data_subset(self, project_id: str, training_data_id: str) -> Dict[str, Any]:
        return self._get(f"projects/{project_id}/data/{training_data_id}/subset")

    def put_training_data_subset(
        self, project_id: str, training_data_id: str, subset: Optional[str]
    ) -> None:
        return self._put(
            f"projects/{project_id}/data/{training_data_id}/subset", json={"subset": subset}
        )
